# coding=utf-8
from collections import OrderedDict
from numpy.linalg import norm
import numpy as np
from ..voronoi.VoronoiEdge import VoronoiEdge
from ..voronoi.VoronoiFace import VoronoiFace
from ...data.AtomImage import AtomImage

class VoronoiCell:
    """Class describing a single cell in a Voronoi tessellation.

    Attributes
    ----------
    atom : Atom
        Atom at center of cell.
    radical : bool
        Whether to add faces using the radical plane Voronoi tessellation
        method.
    faces : array-like
        List of faces in this cell.
    volume : float
        Volume of this cell.

    """
    def __init__(self, atom, radical, faces=None):
        """Function to initialize a Voronoi cell for a specific atom.

        Creates a parallelepiped around the atom defined by its interactions
        with the periodic images across the cube face of the cell, or the
        walls of the cell.

        Parameters
        ----------
        atom : Atom
            Atom at center of cell.
        radical : bool
            Whether to add faces using the radical plane Voronoi tessellation
            method.
        faces : array-like
            List of faces in this cell.
        """

        # Atom at the center of this cell.
        self.atom = atom

        # List of faces.
        self.faces = faces

        # Whether to use radical plane method.
        self.radical = radical

        # Volume of cell. Cached result.
        self.volume = np.nan

    def get_atom(self):
        """Function to get atom at center of this cell.

        Parameters
        ----------

        Returns
        -------
        atom : Atom
        """
        return self.atom

    def get_faces(self):
        """Function to get all faces.

        Parameters
        ----------

        Returns
        -------
        faces : array-like
            List of faces.
        """
        return self.faces

    def n_faces(self):
        """Function to get the number of faces.

        Parameters
        ----------

        Returns
        -------
        type : int
            Number of faces.
        """
        return len(self.faces)

    def get_vertices(self):
        """Function to get all vertices in this cell.

        Parameters
        ----------

        Returns
        -------
        type : array-like
            List of vertices.

        """
        return list(set([face.get_vertices() for face in self.faces]))

    def get_neighbor_types(self):
        """Function to get the number of neighbors of each type.

        Parameters
        ----------

        Returns
        -------
        type : OrderedDict
            Ordered dictionary with atom_type, number of neighbors as key,
            value pairs.
        """
        output = OrderedDict()
        for face in self.faces:
            atom_id = face.get_outside_atom().get_atom_id()
            atom_type = self.atom.get_cell().get_atom(atom_id).get_type()
            if atom_type in output:
                output[atom_type] += 1
            else:
                output[atom_type] = 1

        return output

    def get_neighbors(self):
        """Function to get list of neighboring atoms. Identifies atoms by both
        their ID and image location.

        Parameters
        ----------

        Returns
        -------
        type : array-like
            List of AtomImage's.
        """
        return np.array([face.get_outside_atom() for face in self.faces])

    def get_neighbor_distances(self):
        """Function to get distances between central atom and each neighbor.

        Parameters
        ----------

        Returns
        -------
        type : array-like
            List of distances in cartesian units.
        """
        return np.array([face.get_neighbor_distance() for face in self.faces])

    def get_neighbor_shells(self, cells, index):
        """Function to get list of atoms in all coordination cells.

        Parameters
        ----------
        cells : array-like
            A list of VoronoiCell's of all other atoms (arranged by atom ID).
        index : int
            Index of largest neighbor shell to be considered.

        Returns
        -------
        previous_shells : array-like
            All neighbors in that. A list of list of AtomImage's.

        Raises
        ------
        ValueError
            If index is less than 0.

        """
        if index < 0:
            # Special case: Bad input.
            raise ValueError("Shell index must be >= 0.")
        elif index == 0:
            # Special case: 0th cell (this atom).
            return [OrderedDict({AtomImage(self.atom, [0, 0, 0]):None})]
        else:
            # General case: i-th shell.
            # Get the previous shell.
            previous_shells = self.get_neighbor_shells(cells, index - 1)

            # For each atom in the previous shell, get its neighbors.
            new_shell = OrderedDict()
            for atom in previous_shells[index - 1]:
                # Get its nearest neighbor according to the diagram.
                neighbors = cells[atom.get_atom_id()].get_neighbors()
                # For each of these neighbors, adjust the image coordinates.
                this_image = atom.get_supercell()
                for neighbor in neighbors:
                    # Get the coordinates relative to the central atom.
                    old_image = neighbor.get_supercell() + this_image

                    # Create a new image, and append it to output.
                    new_image = AtomImage(neighbor.get_atom(), old_image)
                    if new_image not in new_shell:
                        new_shell[new_image] = None

            # Remove all images corresponding to the shell inside this one.
            for shell in previous_shells:
                for image in shell:
                    if image in new_shell:
                        new_shell.pop(image)

            # Append new shell to output.
            previous_shells.append(new_shell)

            return previous_shells

    def get_extended_faces(self, cells, index):
        """Function to get faces on the outside of a polyhedron formed by the
        Voronoi cells and its neighbors.

        Note: The coordinates of the center of each face may not be correct.

        Parameters
        ----------
        cells : array-like
            A list of VoronoiCell's of all other atoms (arranged by Atom ID).
        index : int
            Index of largest neighbor shell to be included (0 == faces of
            atom, 1 == faces of atom + 1st nearest neighbor shell, etc.).

        Returns
        -------
        output : array-like
            List of faces (VoronoiFace) on the outside.

        """

        if index == 0:
            # Special case: Index == 0.
            return list(self.faces)

        # Get neighbor list.
        neighbor_list = self.get_neighbor_shells(cells, index)

        # Last shell are atoms on the outside.
        outside_images = neighbor_list[index]

        # Get atoms to consider.
        all_images = OrderedDict({[[image for images in neighbor_list for
                                    image in images]]:None})

        # Get the faces of all outside atoms that do not correspond to an
        # atom on the inside of the polyhedron.
        output = []
        for image in outside_images:
            # Get coordinates of this image w.r.t. central atom.
            image_coord = image.get_supercell()
            # Get its Voronoi cell.
            cell = cells[image.get_atom_id()]
            # Iterate over its faces.
            for face in cell.get_faces():
                # Get the image on the outside of this face.
                relative_image = face.get_outside_atom()
                # Get the coordinates of this image relative to the central
                # atom.
                outside_coord = relative_image.get_supercell() + image_coord

                # Get the actual image.
                actual_image = AtomImage(relative_image.get_atom(),
                                         outside_coord)

                # If this image is not inside of the polyhedron, add the face
                #  to the output.
                if actual_image not in all_images:
                    output.append(face)
        return output

    def get_neighbors_by_walks(self, cells, shell):
        """Function to get the neighbors in a certain shell, along with weights
        defined by face sizes.

        The neighbor shell of an atom is typically defined as all atoms that
        can be reach ed in no fewer than a certain number of steps along the
        a network defined by faces of a Voronoi tessellation. While this works
        in theory, it is complicated by the fact that Voronoi tessellations
        computed numerically often contain very small faces due to numerical
        problems. These atoms with very small faces would be conventionally
        defined as 2nd nearest neighbors but, due to the small faces,
        would be counted as as 1st nearest neighbors.

        To combat this problem, we weighing each neighbor based on the face
        sizes. For a first neighbor shell, the weight of an atom is associated
        with its the fraction of the surface area of the cell related to the
        area of the face corresponding to it. To extend this idea to shells
        past the 1st, we compute the probability of a walker originating from
        this cell ends up at a certain atom after N, non-backtracking walks
        where the probability that it will take a certain step is related to
        the fraction of the face size area.

        For example, let's assume this cell is part of a simple-cubic crystal,
        where each cell has 6, equally-sized faces. The probability that it
        ends up at any first shell neighbor is 1/6. After this step,
        there are 5 possible steps (no backtracking). So, each walk has an
        equal probability of 1/30. There are six atoms that are two steps in
        the same direction (ex: 2,0,0), which can be reached by only 1 path
        and therefore have 1/30 weight. Additionally, there are 12 atoms that
        can be reached by 2 different sets of two orthogonal steps (ex: 1,1,
        0) that, therefore, have a weight of 2/30.

        Parameters
        ----------
        cells : array-like
            A list of VoronoiCell's of all other atoms (arranged by atom ID).
        shell : int
            Which neighbor shell to evaluate.

        Returns
        -------
        output : OrderedDict
            Dictionary containing atoms in that shell as the keys (Atom),
            and their path weights (float) as values. Weights will add up to 1.

        """

        # Gather all possible paths.
        # Initialize with first step.
        paths = []
        # Add in starting point.
        paths.append(([AtomImage(self.atom, [0, 0, 0])], 1.0))
        # Increment until desired shell.
        for step in range(shell):
            # Get the new list.
            new_paths = []
            # For each current path.
            for path in paths:
                # Get last atom in current path.
                last_step = path[0][len(path[0]) - 1]
                # Get each possible step.
                new_steps = OrderedDict()
                # Surface area of possible steps.
                surface_area = 0.0
                for face in cells[last_step.get_atom_id()].get_faces():
                    # Adjust image based on position of last atom.
                    new_supercell = face.get_outside_atom().get_supercell() +\
                                    last_step.get_supercell()

                    # Store the next step.
                    next_step = AtomImage(face.get_outside_atom().get_atom(),
                                          new_supercell)
                    area = face.get_area()
                    surface_area += area
                    new_steps[next_step] = area

                # Eliminate backtracking steps.
                for previous_step in path[0]:
                    if previous_step in new_steps:
                        surface_area -= new_steps.pop(previous_step)

                # Create new paths, making sure to update weights.
                for k,v in new_steps.iteritems():
                    # Increment path.
                    new_path = list(path[0])
                    new_path.append(k)

                    # Increment weight.
                    new_weight = path[1] * v / surface_area

                    # Add it to new_paths.
                    new_paths.append((new_path, new_weight))

            # Update paths.
            paths = new_paths

        # Now that all the paths are gathered, output only the last step and
        # weights of all paths that lead to that step.
        output = OrderedDict()
        for path in paths:
            # Get the last step.
            atom = path[0][len(path[0]) - 1]

            # Update map.
            if atom in output:
                output[atom] += path[1]
            else:
                output[atom] = path[1]

        return output

    def get_neighbor_shell(self, cells, index):
        """Function to get list of atoms in a certain coordination shell.

        A neighbor shell includes all atoms that are a certain number of
        neighbors away that are not in any smaller neighbor shell (e.g. 2nd
        shell neighbors cannot also be 1st nearest neighbors).

        Parameters
        ----------
        cells : array-like
            A list of VoronoiCell's of all other atoms (arranged by Atom ID).
        index : int
            Index of neighbor shell.

        Returns
        -------
        type : array-like
            All neighbors in that. A list of AtomImage's.

        """
        return self.get_neighbor_shells(cells, index)[index]

    def get_num_shared_bonds(self, cell, direction, neighbors):
        """Function to compute number of shared bonds.

        Atoms are defined as bonded if they share a face between their
        Voronoi cells.

        This code determines how many common entries occur between a Voronoi
        cell and a list of neighbor IDs from this cell. Note that in order for
        bonds to be shared, they must connect to the same image of a certain
        atom. That is where the direction parameter comes into play.

        Parameters
        ----------
        cell : VoronoiCell
            Voronoi cell of neighboring atom.
        direction : array-like
            Difference between image of neighboring atom and this cell.
        neighbors : array-like
            A list of AtomImage's containing IDs and Image Positions of all
            neighboring atoms to this cell.

        Returns
        -------
        type : int
            Number of shared neighbors between this cell and the neighboring
            atom.

        """
        n_shared = 0
        for face in cell.get_faces():
            other_cell_n_id = face.get_outside_atom().get_atom_id()
            other_cell_n_image = face.get_outside_atom().get_supercell() + \
                                 direction
            # Make that image.
            other_cell_n = AtomImage(self.atom.get_cell().get_atom(
                other_cell_n_id), other_cell_n_image)

            # Check against every neighbor of this cell.
            for n in neighbors:
                if other_cell_n.__eq__(n):
                    n_shared += 1

        return n_shared

    def get_coordination_shell_shape(self, cells):
        """Function to get the shape of the coordination polyhedron around this
        atom.

        This is determined by counting the number of "bonds" between
        first neighbor shell atoms. Here, atoms are "bonded" if their Voronoi
        cells share a face. This is similar to get_polyhedron_shape. See
        section H of the paper listed in the References section for more
        details.

        Parameters
        ----------
        cells : array-like
            VoronoiCell's of all other atoms (arranged by atom ID).

        References
        ----------
        .. [1] L. Ward, D. Miracle, W. Windl, O. N. Senkov, and K. Flores,
        “Structural evolution and kinetics in Cu-Zr metallic liquids from
        molecular dynamics simulations,” Physical Review B, vol. 88, no. 13,
        Oct. 2013.

        Returns
        -------
        output : OrderedDict
            Ordered dictionary containing the number of bonds, number of
            first shell neighbors with that many bonds as key,value pairs.

        """

        # Get IDs of every neighbor.
        neighbors = [face.get_outside_atom() for face in self.faces]

        # Get number of mutual neighbors for each neighbor.
        output = OrderedDict()
        for n in neighbors:
            n_shared = self.get_num_shared_bonds(cells[n.get_atom_id()],
                                                 n.get_supercell(), neighbors)
            if n_shared in output:
                output[n_shared] += 1
            else:
                output[n_shared] = 1

        return output

    def get_polyhedron_shape(self):
        """Function to compute the polyhedron shape index.

        This shape is defined by the "Voronoi Index", which is defined as the
        number of faces with a certain number of sides. This is often
        expressed as (n_3, n_4, n_5, n_6), where n_x is the number of faces
        with x number of sides.
        http://rspa.royalsocietypublishing.org/cgi/doi/10.1098/rspa.1970.0190
        Finney (1970).

        References
        ----------
        .. [1] J. L. Finney, “Random Packings and the Structure of Simple
        Liquids. II. The Molecular Geometry of Simple Liquids,” Proceedings
        of the Royal Society A: Mathematical, Physical and Engineering
        Sciences, vol. 319, no. 1539, pp. 495–507, Nov. 1970.

        Returns
        -------
        output : OrderedDict
            Ordered dictionary containing number of edges, number of faces
            with that edge count as key,value pairs.
        """
        output = OrderedDict()
        for face in self.faces:
            n_vertices = face.n_vertices()
            if n_vertices in output:
                output[n_vertices] += 1
            else:
                output[n_vertices] = 1

        return output

    def get_volume(self):
        """Function to get volume of this cell

        Returns
        -------
        type : float
            Volume.
        """
        if np.isnan(self.volume):
            self.volume = 0
            atom_center = self.atom.get_position_cartesian()
            for face in self.faces:
                area = face.get_area()
                from_center = face.get_centroid() - atom_center
                n = face.get_normal()
                n /= norm(n)
                h = np.dot(from_center, n)
                self.volume += area * h / 3.0 # Face normal is away from
                # center.

        return self.volume

    def get_surface_area(self):
        """Function to get surface area of cell.

        Returns
        -------
        type : float
            Surface area.
        """
        return sum([face.get_area() for face in self.faces])

    def get_min_max_vertex_distance(self):
        """Function to get the minimum and maximum distance between any two
        vertices.

        Returns
        -------
        min_dist : float
            Minimum distance.
        max_dist : float
            Maximum distance.

        """
        vertices = self.get_vertices()
        l_v = len(vertices)
        min_dist = float("inf")
        max_dist = -min_dist
        for i in range(l_v):
            for j in range(i+1, l_v):
                dist = vertices[i].distance_from(vertices[j])
                min_dist = min(dist, min_dist)
                max_dist = max(dist, max_dist)

        return min_dist, max_dist

    def geometry_is_valid(self):
        """Function to determine whether the geometry of this structure is
        sound.

        Returns
        -------
        type : bool
            True if valid, else False.
        """
        for face in self.faces:
            if not face.is_closed():
                return False

            for f in face.get_neighboring_faces():
                if f not in self.faces:
                    return False
        return True

    def compute_cell(self, image_finder, cutoff):
        """Function to compute cell, given ability to generate images.

        Parameters
        ----------
        image_finder : PairDistanceAnalysis
            Tool to find images within a cutoff.
        cutoff : float
            Initial cutoff. Will be increased if too small.

        Returns
        -------

        Raises
        ------
        Exception
            If cell fails to compute.

        """
        cur_cutoff = cutoff
        n_attempts = 0
        while n_attempts < 4:
            n_attempts += 1
            image_finder.set_cutoff_distance(cur_cutoff)
            # Find all nearby images.
            images = [image[0] for image in
                      image_finder.get_all_neighbors_of_atom(
                          self.atom.get_id())]
            # Compute cell.
            try:
                self.compute_cell_helper(images)
            except Exception:
                cur_cutoff *= 1.5
                continue

            return

        raise Exception("Cell failed to compute.")

    def compute_cell_helper(self, images):
        """Function to compute the Voronoi cell, given list of images.

        Parameters
        ----------
        images : array-like
            List of images to consider, where key is the atomID
            and value is the distance.

        """

        # Clear cached volume.
        self.volume = np.nan

        # Get all possible faces.
        possible_faces = self.compute_faces(images)

        # Get the faces corresponding to the direct polyhedron.
        direct_faces = self.compute_direct_neighbors(possible_faces)

        # Construct direct polyhedron.
        for df in direct_faces:
            try:
                df.assemble_face_from_faces(direct_faces)
            except Exception:
                raise Exception("Direct polyhedron failed to construct.")

        self.faces = list(direct_faces)

        # Get the faces that might actually be direct faces.
        possible_indirect_faces = self.compute_possible_indirect_neighbors(
            possible_faces)

        # Use these faces to compute indirect neighbors.
        for face in possible_indirect_faces:
            self.compute_intersection(face)

    def compute_faces(self, images):
        """Function to compute the center of the face corresponding to each
        neighbor.

        Parameters
        ----------
        images : array-like
            List of all images of this atom.

        Returns
        -------
        output : array-like
            List of faces.

        Raises
        ------
        RuntimeError
            If it fails to create face.
        """

        # Generate faces.
        output = []
        for image in images:
            # Check if the image is this atom.
            if image.get_atom_id() == self.atom.get_id() and np.array_equal(
                    image.get_supercell(), [0, 0, 0]):
                # If so, skip.
                continue

            # Make the appropriate face.
            try:
                output.append(VoronoiFace(self.atom, image, self.radical))
            except Exception:
                raise RuntimeError("Failure to create face.")

        return output

    def compute_direct_neighbors(self, faces):
        """Function to compute list of neighbors for the direct polyhedron
        and those that could contribute to the Voronoi polyhedron.

        See page 85 of Ref. [1].

        Parameters
        ----------
        faces : array-like
            List of faces to consider. After this operation,
            any faces that are on the direct polyhedron will be removed and the
            list will be sorted by distance from the center.

        References
        ----------
        .. [1] W. Brostow, J.-P. Dussault, and B. L. Fox, “Construction of
        Voronoi polyhedra,” Journal of Computational Physics, vol. 29, no. 1,
        pp. 81–92, Oct. 1978.

        Returns
        -------
        direct_faces : array-like
            List of faces that are on the direct polyhedron.

        """

        # Sort distance from face to the center.
        faces.sort(cmp=self.compare_faces)

        # The closest face is on the direct polyhedron.
        direct_faces = []
        direct_faces.append(faces.pop(0))

        # Now loop through all faces to find those whose centers are not
        # outside any other face that is on the direct polyhedron.

        to_remove = []
        for face in faces:
            is_inside = True
            for df in direct_faces:
                d = df.position_relative_to_face(face.get_face_center())
                if d >= 0:
                    is_inside = False
                    break

            if is_inside:
                direct_faces.append(face)
                to_remove.append(face)
        for face in to_remove:
            faces.remove(face)
        return direct_faces

    def compare_faces(self, a, b):
        """Function to compare two Voronoi faces.

        Parameters
        ----------
        a : VoronoiFace
            Face a.
        b : VoronoiFace
            Face b.

        Returns
        -------
        type : int
            1 if a < b, +1 if a > b, else 0.

        """
        d1 = a.get_face_distance()
        d2 = b.get_face_distance()

        # Have to check if they are really close. If they are, return 0.
        if (d1 - d2) ** 2 < 1e-30:
            return 0
        elif d1 < d2:
            return -1
        else:
            return +1

    def compute_possible_indirect_neighbors(self, possible_faces):
        """Function to get list of faces that might possibly be indirect.

        You must already have computed direct faces and all vertices of the
        direct polyhedron.

        Parameters
        ----------
        possible_faces : array-like
            List of all possible non-direct faces sorted in ascending order.
            This probably has already been computed by compute_direct_neighbors.

        Returns
        -------
        possible_indirect_faces : array-like
            List of faces that might actually be indirect neighbors. Here,
            we find if the face distance is less than the farthest vertex on the
            direct polyhedron.

        """

        # Get list of faces that might be indirect neighbors.
        max_vertex_distance = max([v.get_distance_from_center() for face in
                                   self.faces for v in face.get_vertices()])

        possible_indirect_faces = []
        for face in possible_faces:
            if face.get_face_distance() < max_vertex_distance:
                possible_indirect_faces.append(face)
            else:
                # possible faces is sorted by get_direct_faces.
                break

        return possible_indirect_faces

    def compute_intersection(self, new_face):
        """Function to compute intersection between a potential face and this
        cell.

        Parameters
        ----------
        new_face : VoronoiFace
            Potential new face.

        Returns
        -------
        type : bool
            Whether it intersected this cell.

        Raises
        ------
        Exception
            If not enough edges were formed.
            If geometry is invalid.
        """

        # Clear cached result.
        self.volume = np.nan

        # Determine whether any faces intersect this new new_face.
        no_intersection = True
        for cur_face in self.faces:
            for v in cur_face.get_vertices():
                d = new_face.position_relative_to_face(v.get_position())
                if d > 0:
                    no_intersection = False
                    break

        if no_intersection:
            return False

        # If it does, store the old face information.
        previous_state = OrderedDict({face: face.get_edges() for face in
                                     self.faces})

        # Attempt to perform intersections.
        try:
            new_edges = []
            to_remove = []
            for c_face in self.faces:
                new_edge = c_face.compute_intersection(new_face)
                if c_face.n_edges() < 3:
                    to_remove.append(c_face)
                else:
                    if new_edge is not None:
                        new_edges.append(new_edge)
            for f in to_remove:
                self.faces.remove(f)
            # Check if we have enough edges.
            if len(new_edges) < 3:
                raise Exception("Not enough edges were formed.")

            # Assemble new face and add it to list of faces.
            new_face.assemble_face_from_edges(new_edges)
            self.faces.append(new_face)

            # Check if geometry is valid.
            if not self.geometry_is_valid():
                raise Exception("Geometry is invalid.")
            return True
        except Exception:
            # Restore previous state.
            self.faces = []
            for face in previous_state:
                face.reset_edges(previous_state[face])
                self.faces.append(face)
            return False

    def remove_face(self, to_remove):
        """Function to remove a face from this cell.

        Parameters
        ----------
        to_remove : VoronoiFace
            Face to remove.

        Raises
        ------
        RuntimeError
            If no such face exists.
        """
        if to_remove not in self.faces:
            raise RuntimeError("No such face exists.")

        self.faces.remove(to_remove)
        self.volume = np.nan

        # Find all faces that are currently in contact with this face.
        contacting_faces = to_remove.get_neighboring_faces()
        for face_to_build in contacting_faces:
            cur_edges = face_to_build.get_edges()

            # Compute edges corresponding to the "contacting faces".
            for face in contacting_faces:
                if face.__eq__(face_to_build):
                    continue
                new_edge = VoronoiEdge(face_to_build, face)
                cur_edges.append(new_edge)

            # Remove the edge corresponding to the face being removed.
            for e in cur_edges:
                if e.get_intersecting_face().__eq__(to_remove):
                    cur_edges.remove(e)
            face_to_build.assemble_face_from_edges(cur_edges)


# coding=utf-8
import types
import pandas as pd
from ....data.materials.CrystalStructureEntry import CrystalStructureEntry
from ....data.materials.util.LookUpData import LookUpData
from ....vassal.analysis.APRDFAnalysis import APRDFAnalysis

class APRDFAttributeGenerator:
    """Class to generate attributes based on the Atomic Property Weighted
    Radial Distribution Function (AP-RDF) approach of Fernandez et al. [1].

    User can specify the cutoff distance for the AP-RDF, the number of points
    to evaluate it, the smoothing factors for the RDF peaks, and the
    properties used for weighting. The recommended values of these
    parameters have yet to be determined, please contact Logan Ward or the
    authors of this paper if you have questions or ideas for these parameters.

    Attributes
    ----------
    cut_off_distance : float
        Cutoff distance for RDF.
    num_points : int
        Number of points to evaluate.
    smooth_parameter : float
        Smoothing parameter for AP-RDF.
    elemental_properties : list
        Elemental properties to be associated with this class for the
        generation of features.

    References
    ----------
    .. [1] M. Fernandez, N. R. Trefiak, and T. K. Woo, "Atomic Property
    Weighted Radial Distribution Functions Descriptors of Metal–Organic
    Frameworks for the Prediction of Gas Uptake Capacity," The Journal of
    Physical Chemistry C, vol. 117, no. 27, pp. 14095--14105, Jul. 2013.

    """

    def __init__(self):
        """Function to create instance and initialize fields.

        """

        # Cutoff distance for RDF.
        self.cut_off_distance = 10.0

        # Number of points to evaluate.
        self.num_points = 6

        # Smoothing parameter for AP-RDF.
        self.smooth_parameter = 4.0

        # List of elemental properties to use for weighting.
        self.elemental_properties = []

    def set_smoothing_parameter(self, b):
        """Function to set smoothing factor used when computing PRDF.

        Parameters
        ----------
        b : float
            Smoothing factor.

        """

        self.smooth_parameter = b

    def set_cut_off_distance(self, d):
        """Function to set cut off distance used when computing PRDF.

        Parameters
        ----------
        d : float
            Cut off distance.

        """

        self.cut_off_distance = d

    def set_num_points(self, num_points):
        """Function to set the number of points at which to evaluate AP-RDF.

        Parameters
        ----------
        num_points : int
            Desired number of windows.

        """

        self.num_points = num_points

    def add_elemental_property(self, property_name):
        """Function to add an elemental property to `self.elemental_properties`
        in order to be used to compute features.

        Parameters
        ----------
        property : str
            Property to be added.

        """

        if property_name not in self.elemental_properties:
            self.elemental_properties.append(property_name)

    def add_elemental_properties(self, properties):
        """Function to provide a list of elemental properties to be used to
        compute features.

        Parameters
        ----------
        properties : array-like
            Properties to be included. A list of strings containing property
            names.

        """

        for prop in properties:
            self.add_elemental_property(prop)

    def clear_elemental_properties(self):
        """Function to clear the list of elemental properties.

        """

        self.elemental_properties = []

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Crystal structures for which features are to be generated. A list
            of CrystalStructureEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CrystalStructureEntry instances.

        """

        # Initialize list of feature values for pandas data frame.
        feat_values = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")
        elif (entries and not isinstance(entries[0], CrystalStructureEntry)):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")

        # Create tool to compute AP-RDF.
        tool = APRDFAnalysis()
        tool.set_cut_off_distance(self.cut_off_distance)
        tool.set_n_windows(self.num_points)
        tool.set_smoothing_factor(self.smooth_parameter)

        # Insert header names here.
        eval_dist = tool.get_evaluation_distances()
        feat_headers = ["APRDF_{:s}_R={:f}_B={:f}".format(prop, dist,
                        self.smooth_parameter) for prop in
                        self.elemental_properties for dist in eval_dist]

        # Loop through each entry, compute attributes.
        for entry in entries:
            # Get the structure.
            structure = entry.get_structure()

            # Prepare the APRDF tool.
            tool.analyze_structure(structure)

            # Loop through each property.
            for prop in self.elemental_properties:
                prop_lookup = LookUpData.load_property(prop)
                atom_prop = [prop_lookup[LookUpData.element_names.index(
                    structure.get_type_name(t))] for t in range(
                    structure.n_types())]

                # Compute the APRDF.
                ap_rdf = tool.compute_APRDF(atom_prop)

                feat_values.append(ap_rdf)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import numpy as np
from numpy.linalg import norm
import sys

class Line:
    """Class to represent lines in a three dimensional space.

    Documentation obtained from http://commons.apache.org/proper/commons-math
    /javadocs/api-3.3/org/apache/commons/math3/geometry/euclidean/threed/Line
    .html
    Each oriented line is intrinsically associated with an abscissa which is
    a coordinate on the line. The point at abscissa 0 is the orthogonal
    projection of the origin on the line, another equivalent way to express
    this is to say that it is the point of the line which is closest to the
    origin. Abscissa increases in the line direction.

    Attributes
    ----------
    direction : array-like
        Line direction.
    zero : array-like
        Line point closest to the origin.
    tolerance : float
        Tolerance below which points are considered identical.

    """

    def __init__(self, p1=None, p2=None, tolerance=None, l=None):
        """Constructor to create instance of the Line object.

        Parameters
        ----------
        p1 : array-like
            First point belonging to the Line (this can be any point).
        p2 : array-like
            Second point belonging to the Line (this can be any point,
            different from p1).
        tolerance : float
            Tolerance below which points are considered identical.
        l : Line
            Line to copy.
        """

        if l is None:
            p1_arr = np.array(p1, dtype=float)
            p2_arr = np.array(p2, dtype=float)

            delta = p2_arr - p1_arr
            norm2 = delta[0] ** 2 + delta[1] ** 2 + delta[2] ** 2
            norm1 = np.math.sqrt(norm2)
            if norm2 == 0:
                raise Exception("Norm is zero!")

            # Line direction.
            self.direction = delta / norm1

            # Line point closest to the origin.
            self.zero = p1_arr - np.dot(p1_arr, delta) / norm2

            # Tolerance below which points are considered identical.
            self.tolerance = 1e-10 if tolerance is None else tolerance
        else:
            # Tolerance below which points are considered identical.
            self.tolerance = l.tolerance

            # Line direction.
            self.direction = l.direction

            # Line point closest to the origin.
            self.zero = l.zero

    def set_tolerance(self, tol):
        """Function to set the tolerance.

        Parameters
        ----------
        tol : float
            Desired tolerance.

        """
        self.tolerance = tol

    def get_tolerance(self):
        """Function to get the tolerance.

        Returns
        -------
        output : float
            Tolerance.

        """
        return self.tolerance

    def revert(self):
        """Function to revert the direction of the current line.

        Returns
        -------
        output : Line
            A new Line instance with the direction reversed.
        """
        reverted = Line(l=self)
        reverted.direction *= -1.0
        return reverted

    def get_direction(self):
        """Function to get the direction of the line.

        Returns
        -------
        output : array-like
            Direction as a numpy array.
        """
        return self.direction

    def get_origin(self):
        """Get the line point closest to the origin.

        Returns
        -------
        output : array-like
            Point as a numpy array.
        """
        return self.zero

    def get_abscissa(self, p):
        """Function to get the abscissa of a point with respect to a line.

        The abscissa is 0 if the projection of the point and the projection of
        the frame origin on the line are the same point.

        Parameters
        ----------
        p : array-like
            Desired point.

        Returns
        -------
        output : float
            Abscissa of the point.

        """
        return np.dot(p - self.zero, self.direction)

    def point_at(self, abscissa):
        """Function to get one point from the line.

        Parameters
        ----------
        abscissa : float
            Desired abscissa for the point/

        Returns
        -------
        output : array-like
            One point belonging to the line, at specified abscissa.

        """
        return self.zero + abscissa * self.direction

    def contains(self, p):
        """Function to check if the instance contains a point.

        Parameters
        ----------
        p : array-like
            Point to check.

        Returns
        -------
        output : bool
            True if p belongs to the line, else False.

        """
        return self.distance(p=p) < self.tolerance

    def distance(self, p=None, l=None):
        """Function to compute the distance between the instance and a point or
        the shortest distance between the instance and another line.

        Parameters
        ----------
        p : array-like
            Point to compute distance between. (Default value = None)
        l : Line
            Line to compute distance between. (Default value = None)

        Returns
        -------
        output : float
            Distance.

        """
        if l is None:
            d = p - self.zero
            n = np.zeros(3)
            try:
                n = d - np.dot(d, self.direction) * self.direction
            except RuntimeWarning:
                print d, self.direction
            return norm(n)
        else:
            normal = np.cross(self.direction, l.direction)
            n = norm(normal)
            if n < sys.float_info.min:
                # Lines are parallel.
                return self.distance(p=l.zero)
            offset = np.dot(l.zero - self.zero, normal) / n
            return np.abs(offset)

    def distance_sq(self, p):
        """Function to compute the distance squared between the instance and a
        point.

        Parameters
        ----------
        p : array-like
            Point to compute distance between.

        Returns
        -------
        output : float
            Distance.

        """
        d = p - self.zero
        n = d - np.dot(d, self.direction) * self.direction
        return n[0] ** 2 + n[1] ** 2 + n[2] ** 2

    def closest_point(self, l):
        """Function to the compute the point of the instance closest to another
        line.

        Parameters
        ----------
        l : Line
            Line to check against the instance.

        Returns
        -------
        output : array-like
            Point of the instance closest to another line.

        """
        cos = np.dot(self.direction, l.direction)
        n = 1 - cos ** 2
        if n < sys.float_info.epsilon:
            # Lines are parallel.
            return self.zero

        d0 = l.zero - self.zero
        a = np.dot(d0, self.direction)
        b = np.dot(d0, l.direction)
        return self.zero + self.direction * ( a - b * cos) / n

    def intersection(self, l):
        """Function to compute the intersection point of the instance and
        another line.

        Parameters
        ----------
        l : Line
            Other line.

        Returns
        -------
        output : array-like
            Intersection point of the instance and the other line,
            else None.

        """
        closest = self.closest_point(l)
        return closest if l.contains(closest) else None


import numpy as np
from numpy.linalg import norm
from scipy.linalg import lu_factor, lu_solve

class VectorCombinationComputer:
    """Class to find all combinations of 3 vectors that are shorter than a
    certain length.

    Attributes
    ----------
    input_vectors : array-like
        Vectors to be added.
    cutoff_dist_sq :  float
        Square of cutoff distance.
    include_zero : bool
        Whether to include the zero vector in the list.
    super_cells : list
        x, y, z coordinates of each vector shorter than cutoff.
    vectors : list
        All vectors shorter then cutoff.

    """
    def __init__(self, in_vectors, cutoff_distance, include_zero=True):
        """Function to create the tool to compute all combinations of input
        vectors shorter than cutoff distance.

        Parameters
        ----------
        in_vectors : array-like
            Vectors to be combined. Must be exactly 3 linearly-independent
            vectors.
        cutoff_distance :  float
            Desired cutoff distance.
        include_zero : bool
            Whether to include the zero vector in the list.

        Raises
        ------
        ValueError
            If length of in_vectors is not 3.
        """

        if len(in_vectors) != 3:
            raise ValueError("Expecting exactly three vectors.")

        # Vectors to be added.
        self.input_vectors = list(in_vectors)

        # Square of cutoff distance.
        self.cutoff_distance_sq = cutoff_distance ** 2

        # Whether to include the zero vector in the list.
        self.include_zero = include_zero

        # x, y, z coordinates of each vector shorter than cutoff.
        self.super_cells = []

        # All vectors shorter then cutoff.
        self.vectors = []

        self.get_all_vectors()

    def compute_vector(self, x):
        """Function to compute a certain combination of vectors stored in
        this array.

        Parameters
        ----------
        x : array-like

        Returns
        -------
        output : array-like
            Combination.
        """
        i_v = np.array(self.input_vectors, dtype=float)
        return np.array([x.dot(y) for y in i_v.T])


    def get_all_vectors(self):
        """Function to compute all vectors shorter than cutoff distance.
        """

        # Create a matrix of basis vectors.
        basis = np.array(self.input_vectors, dtype=float).T

        # Create ability to invert it.
        det_basis = np.linalg.det(basis)


        if det_basis == 0 or det_basis < 1e-14:
            raise RuntimeError("Vectors are not linearly independent.")

        fac = lu_factor(basis)

        # Compute range of each variable.
        cutoff_distance = np.math.sqrt(self.cutoff_distance_sq)
        step_range = []

        for i in range(3):
            max_disp = 0.0
            for j in range(3):
                max_disp += np.dot(self.input_vectors[i], self.input_vectors[
                    j]) / norm(self.input_vectors[i])
            step_range.append(int(np.math.ceil(max_disp / cutoff_distance)) + 1)

        # Ensure that we have sufficient range to get the cutoff distance
        # away from the origin by checking that we have large enough range to
        #  access a point cutoff distance away along the direction of xy,
        # xz and yz cross products.
        for dir in range(3):
            point = np.cross(self.input_vectors[dir], self.input_vectors[(dir
                                    + 1) % 3])
            point = point * cutoff_distance / norm(point)
            sln = lu_solve(fac, point)
            step_range = [max(step_range[i], int(np.math.ceil(abs(sln[i]))))
                          for i in range(3)]

        # Create the initial vector.
        for x in range(-step_range[0], 1 + step_range[0]):
            for y in range(-step_range[1], 1 + step_range[1]):
                for z in range(-step_range[2], 1 + step_range[2]):
                    a = np.array([x, y, z])
                    l = self.compute_vector(a)
                    dist_sq = l[0] ** 2 + l[1] ** 2 +  l[2] ** 2
                    if dist_sq <= self.cutoff_distance_sq:
                        if not self.include_zero and x == 0 and y == 0 and z \
                                == 0:
                            continue
                        self.super_cells.append(a)
                        self.vectors.append(l)

    def get_vectors(self):
        """Function to get the list of all vectors shorter than cutoff.

        Returns
        -------
        output : array-like
            List of vectors.

        """
        return list(self.vectors)

    def get_supercell_coordinates(self):
        """Function to get the list of all image coordinates of vectors.

        Returns
        -------
        output : array-like
            List of supercell coordinates.
        """
        return list(self.super_cells)


from numpy.linalg import norm
import numpy as np

class VoronoiVertex:
    """Class for a vertex in a Voronoi tessellation.

    Attributes
    ----------
    position : array-like
        Position of vertex.
    distance : float
        Distance from cell center.
    next_edge : VoronoiEdge
        Edge that is after this vertex.
    previous_edge : VoronoiEdge
        Edge that is before this vertex.
    """
    def __init__(self, inside_atom=None, position=None, edge1=None, edge2=None):
        """Function to initialize a Voronoi vertex. Can specify either inside
        atom and position or two edges.

        Parameters
        ----------
        inside_atom : Atom
            Atom on "inside" of this face.
        position : array-like
            Position of vertex.
        edge1 : VoronoiEdge
            Edge 1.
        edge2 : VoronoiEdge
            Edge 2.

        Raises
        ------
        ValueError
            Position should be a numpy array!.
        """

        if position is not None and not isinstance(position, np.ndarray):
            raise ValueError("Position should be a numpy array!.")

        in_atom = inside_atom
        pos = position

        # Edge that is before this vertex.
        self.previous_edge = edge1

        # Edge that is after this vertex.
        self.next_edge = edge2

        if in_atom is None and pos is None:
            in_atom = edge1.get_edge_face().get_inside_atom()
            pos = edge1.get_line().intersection(edge2.get_line())

        # Position of vertex.
        self.position = pos

        # Distance from cell center.
        self.distance = norm(self.position - in_atom.get_position_cartesian())

        if inside_atom is None and position is None:
            # Store next and previous edges.
            if edge1.is_ccw(edge2=edge2):
                self.previous_edge = edge1
                self.next_edge = edge2
            else:
                self.previous_edge = edge2
                self.next_edge = edge1

    @classmethod
    def get_centroid(self, points):
        """Function to compute the centroid of a group of vertices.

        Parameters
        ----------
        points : array-like
            Points to be considered.

        Returns
        -------
        center : array-like
            Centroid.

        """
        center = np.zeros(3)
        for p in points:
            center += p.get_position()

        center /= len(points)
        return center

    def distance_from(self, vertex):
        """Function to compute distance between vertices.

        Parameters
        ----------
        vertex : VoronoiVertex
            Other vertex.

        Returns
        -------
        output : float
            Distance between them.

        """
        return norm(self.position - vertex.position)


    def get_distance_from_center(self):
        """Function to get the distance from the center of the cell.

        Returns
        -------
        output : float
            Distance.
        """
        return self.distance

    def get_position(self):
        """Function to get the position of this vertex.

        Parameters
        ----------

        Returns
        -------
        output : array-like
            Position of this vertex.
        """
        return self.position

    def __str__(self):
        """Function to print the position of the vertex.

        Returns
        -------
        output : str
            String containing the position of the vertex.
        """
        return str(self.position)

    def __hash__(self):
        """Function to compute the hash value of the vertex.

        Returns
        -------
        value : int
            Hash value.
        """
        return 3 + hash(self.position)

    def __eq__(self, other):
        """Function to check the instance is equal to another vertex.

        Parameters
        ----------
        other : VoronoiVertex
            Other vertex to check.

        Returns
        -------
        value : bool
            True if equal, else False.
        """
        if isinstance(other, VoronoiVertex):
            return other.previous_edge.__eq__(self.previous_edge) and \
                   other.next_edge.__eq__(self.next_edge)
        return False

    def __cmp__(self, other):
        """Function to compare the instance with another vertex.

        Parameters
        ----------
        other : VoronoiVertex
            Other vertex to compare.

        Returns
        -------
        value : int
            -1 if instance < other, +1 if instance > other, else 0.
        """
        if self.previous_edge.__eq__(other.previous_edge):
            return self.next_edge.__cmp__(other.next_edge)
        else:
            return self.previous_edge.__cmp__(other.previous_edge)

    def get_next_edge(self):
        """Function to get the edge after this vertex.

        Returns
        -------
        next_edge : VoronoiEdge
            Edge that is after this vertex.
        """
        return self.next_edge

    def get_previous_edge(self):
        """Function to get the edge before this vertex.

        Returns
        -------
        previous_edge : VoronoiEdge
            Edge that is before this vertex.

        """
        return self.previous_edge

    def is_on_edge(self, e):
        """Function to check if a vertex is on a certain edge.

        Parameters
        ----------
        e : VoronoiEdge
            Edge to check.

        Returns
        -------
        output : bool
            True if instance is on edge, else False.

        """
        return self.previous_edge.__eq__(e) or self.next_edge.__eq(e)


import unittest
import pandas as pd

from cheml.preprocessing import ConstantColumns


# dummy data
df = pd.DataFrame([1,2,3])
df[1] = ['a']*3
df[2] = [7]*3
df[3] = [5.432423]*3


class TestConstantColumns(unittest.TestCase):
    def test_fit_transform(self):
        cc = ConstantColumns()
        f = cc.fit_transform(df)
        self.assertEqual(f.shape, (3, 1))
        self.assertEqual(0, f.columns[0])

    def test_transform(self):
        cc = ConstantColumns()
        f = cc.fit_transform(df)
        ff = cc.transform(df)
        self.assertEqual(3, (ff==f).sum()[0])

if __name__== '__main__':
    unittest.main()




import unittest
from ...utility.EqualSumCombinations import EqualSumCombinations

class testEqualSumCombinations(unittest.TestCase):
    def test_get_combinations(self):
        x = EqualSumCombinations(2, 2)
        self.assertTrue(len(x.get_combinations(2, 2)) == 3)
        x = EqualSumCombinations(2, 3)
        self.assertTrue(len(x.get_combinations(2, 3)) == 6)


import unittest
import os
import pkg_resources

from cheml.initialization import XYZreader

DATA_PATH = pkg_resources.resource_filename('cheml', os.path.join('tests', 'data'))

print 'path:', DATA_PATH

class TestXYZreader(unittest.TestCase):
    def test_string(self):
        reader= XYZreader(path_pattern='[2-3]/*.opt.xyz',
                                           path_root = DATA_PATH,
                                           reader='manual',
                                           skip_lines=[0, 0])
        molecules = reader.read()
        self.assertEqual(len(molecules) , 2)
        self.assertEqual(len(molecules[1]['mol']), 38)

    def test_list(self):
        reader = XYZreader(path_pattern = ['[2-3]/*.opt.xyz', '[1-2][1-2]/*.opt.xyz'],
                                           path_root = DATA_PATH,
                                           reader='manual',
                                           skip_lines = [0, 0])
        molecules = reader.read()
        self.assertEqual(len(molecules) , 4)
        self.assertEqual(len(molecules[1]['mol']), 38)
        self.assertEqual(len(molecules[4]['mol']), 41)


if __name__== '__main__':
    unittest.main()






import types
import numpy as np
import pandas as pd
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.LookUpData import LookUpData

class ElementPairPropertyAttributeGenerator:
    """Class to generate attributes based on the properties of constituent
    binary systems.

    Computes the minimum, maximum and range of all pairs in
    the material, and the fraction-weighted mean and variance of all pairs.
    Variance is defined as the mean absolute deviation from the mean over all
    pairs. If an entry has only one element, the value of NaN is used for all
    attributes.

    Attributes
    ----------
    elemental_pair_properties : list
        Elemental properties to be associated with this class for the generation of
        features.
    pair_lookup-data : dict
        Dictionary containing the property name as the key and a list of floats as
        the value.

    """

    elemental_pair_properties = []
    pair_lookup_data = {}

    def load_pair_lookup_data(self):
        """Function to load the property values into `self.lookup_data` for the
        computation of features.

        """

        self.pair_lookup_data = LookUpData.load_pair_properties(
            self.elemental_pair_properties)

    def add_elemental_pair_property(self, property):
        """Function to add an elemental pair property to be used to compute
        features.

        Parameters
        ----------
        property : str
            Property to be added.

        """

        if property not in self.elemental_pair_properties:
            self.elemental_pair_properties.append(property)

    def add_elemental_pair_properties(self, properties):
        """Function to provide a list of elemental pair properties to be used to
        compute features.

        Parameters
        ----------
        properties : array-like
            Properties to be included. A list of strings containing property
            names.

        """

        for prop in properties:
            self.add_elemental_pair_property(prop)

    def remove_elemental_pair_property(self, property):
        """Function to remove an elemental pair property from the list of
        elemental properties.

        Parameters
        ----------
        property : str
            Property to be removed.

        """

        if property in self.elemental_pair_properties:
            self.elemental_pair_properties.remove(property)

    def remove_elemental_pair_properties(self, properties):
        """Function to remove a list of elemental pair properties from the
        list of elemental properties.

        Parameters
        ----------
        properties : array-like
            Properties to be removed. A list of strings containing property
            names.

        """

        for prop in properties:
            self.remove_elemental_pair_property(prop)

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Compositions for which features are to be generated. A list of
            CompositionEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If no elemental properties are set.
            If input is not of type list.
            If items in the list are not CompositionEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Make sure that there is at least one elemental pair property provided.
        if not self.elemental_pair_properties:
            raise ValueError("No elemental property is set. Add at least one "
                             "property to compute meaningful descriptors.")

        # If the dictionary containing the property values is empty,
        # load values into it.
        if not self.pair_lookup_data:
            self.load_pair_lookup_data()

        # Raise exception if input argument is not of type list of
        # Composition Entry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")

        # Insert header names here.
        n_statistics = 5
        for prop in self.elemental_pair_properties:
            feat_headers.append("binary_max_" + prop)
            feat_headers.append("binary_min_" + prop)
            feat_headers.append("binary_range_" + prop)
            feat_headers.append("binary_mean_" + prop)
            feat_headers.append("binary_variance_" + prop)

        for entry in entries:
            tmp_list = []
            elem_ids = entry.get_element_ids()
            elem_fractions = entry.get_element_fractions()

            if len(elem_fractions) == 1:
                for i in range(n_statistics):
                    tmp_list.append(np.nan)
                feat_values.append(tmp_list)
                continue

            pair_weights = []
            for i in range(len(elem_fractions)):
                for j in range(i):
                    pair_weights.append(elem_fractions[i]*elem_fractions[j])

            total_sum = sum(pair_weights)
            for i in range(len(pair_weights)):
                pair_weights[i] /= total_sum

            # Look up values for each pair property.
            for prop in self.elemental_pair_properties:
                tmp_prop = []

                for i in range(len(elem_fractions)):
                    e_i = elem_ids[i]
                    for j in range(i):
                        e_j = elem_ids[j]
                        idx_1 = max(e_i, e_j)
                        idx_2 = min(e_i, e_j)
                        tmp_prop.append(self.pair_lookup_data[prop][idx_1][
                                            idx_2])

                max_ = max(tmp_prop)
                min_= min(tmp_prop)
                range_ = max_ - min_
                mean_ = np.average(tmp_prop, weights=pair_weights)
                variance_ = np.average([abs(x - mean_) for x in tmp_prop],
                                      weights=pair_weights)
                tmp_list.append(max_)
                tmp_list.append(min_)
                tmp_list.append(range_)
                tmp_list.append(mean_)
                tmp_list.append(variance_)


            feat_values.append(tmp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import numpy as np
from cheml.wrappers.database.containers import Input, Output, Parameter, req, regression_types, cv_types

class mask(object):
    task = ""
    subtask = ""
    host = ""
    function = ""
    requirements = (req(0), req(2))
    documentation = ""

    class Inputs:
        df = Input('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's X class",())

    class Outputs:
        df = Output('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's X class",())

    class WParameters:
        """ Wrapper parameters"""
        func_method = Parameter('func_method', None, 'string')

    class Fparameters:
        """ Function parameters"""
        #
        pass

class PolynomialFeatures(object):
    task = 'Prepare'
    subtask = 'feature representation'
    host = 'sklearn'
    function = 'PolynomialFeatures'
    module = ('sklearn', 'preprocessing')
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.PolynomialFeatures.html#sklearn.preprocessing.PolynomialFeatures"

    class Inputs:
        df = Input('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's PolynomialFeatures class",
                    ["<class 'sklearn.preprocessing.data.PolynomialFeatures'>"])

    class Outputs:
        df = Output('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's PolynomialFeatures class",
                    ["<class 'sklearn.preprocessing.data.PolynomialFeatures'>"])

    class WParameters:
        """ Wrapper parameters"""
        func_method = Parameter('func_method', None, 'string',
                                description = "fit_transform: always make a new api; transform: must receive an api; None: only make a new api ",
                                options=('fit_transform', 'transform', None))

    class Fparameters:
        """ Function parameters"""
        #{'include_bias': True, 'interaction_only': False, 'degree': 2}
        pass

class Binarizer(object):
    task = 'Prepare'
    subtask = 'feature representation'
    host = 'sklearn'
    function = 'Binarizer'
    module = ('sklearn', 'preprocessing')
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Binarizer.html#sklearn.preprocessing.Binarizer"

    class Inputs:
        df = Input('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's Binarizer class",
                    ["<class 'sklearn.preprocessing.data.Binarizer'>"])

    class Outputs:
        df = Output('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's Binarizer class",
                    ["<class 'sklearn.preprocessing.data.Binarizer'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                description="fit_transform: always make a new api; transform: must receive an api; None: only make a new api ",
                                options=('fit_transform', 'transform', None))
    class Fparameters:
        #{'threshold': 0.0, 'copy': True}
        pass

class OneHotEncoder(object):
    task = 'Prepare'
    subtask = 'feature representation'
    host = 'sklearn'
    function = 'OneHotEncoder'
    module = ('sklearn', 'preprocessing')
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html#sklearn.preprocessing.OneHotEncoder"

    class Inputs:
        df = Input('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's OneHotEncoder class",
                    ["<class 'sklearn.preprocessing.data.OneHotEncoder'>"])

    class Outputs:
        df = Output('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's OneHotEncoder class",
                    ["<class 'sklearn.preprocessing.data.OneHotEncoder'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                description="fit_transform: always make a new api; transform: must receive an api; None: only make a new api ",
                                options=('fit_transform', 'transform', None))
    class Fparameters:
        # {'dtype': np.float, 'handle_unknown': 'error', 'sparse': True, 'categorical_features': 'all', 'n_values': 'auto'}
        pass

class Imputer(object):
    task = 'Prepare'
    subtask = 'preprocessor'
    host = 'sklearn'
    function = 'Imputer'
    module = ('sklearn', 'preprocessing')
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Imputer.html#sklearn.preprocessing.Imputer"

    class Inputs:
        df = Input('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's Imputer class",
                    ["<class 'sklearn.preprocessing.imputation.Imputer'>"])

    class Outputs:
        df = Output('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's Imputer class",
                    ["<class 'sklearn.preprocessing.imputation.Imputer'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                description="fit_transform: always make a new api; transform: must receive an api; None: only make a new api ",
                                options=('fit_transform', 'transform', None))

    class Fparameters:
        # {'copy': True, 'strategy': 'mean', 'axis': 0, 'verbose': 0, 'missing_values': 'NaN'}
        pass

class StandardScaler(object):
    task = 'Prepare'
    subtask = 'scale'
    host = 'sklearn'
    function = 'StandardScaler'
    module = ('sklearn', 'preprocessing')
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html#sklearn.preprocessing.StandardScaler"

    class Inputs:
        df = Input('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's StandardScaler class",
                    ["<class 'sklearn.preprocessing.data.StandardScaler'>"])

    class Outputs:
        df = Output('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's StandardScaler class",
                    ["<class 'sklearn.preprocessing.data.StandardScaler'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                description="fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                                options = ('fit_transform','transform','inverse_transform',None))

    class Fparameters:
        # {'copy': True, 'with_mean': True, 'with_std': True}
        pass

class MinMaxScaler(object):
    task = 'Prepare'
    subtask = 'scale'
    host = 'sklearn'
    function = 'MinMaxScaler'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html#sklearn.preprocessing.MinMaxScaler"

    class Inputs:
        df = Input('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's MinMaxScaler class",
                    ["<class 'sklearn.preprocessing.data.MinMaxScaler'>"])

    class Outputs:
        df = Output('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's MinMaxScaler class",
                    ["<class 'sklearn.preprocessing.data.MinMaxScaler'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                description="fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                                options = ('fit_transform','transform','inverse_transform',None))

    class Fparameters:
        # {'copy': True, 'feature_range': (0, 1)}
        pass

class MaxAbsScaler(object):
    task = 'Prepare'
    subtask = 'scale'
    host = 'sklearn'
    function = 'MaxAbsScaler'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MaxAbsScaler.html#sklearn.preprocessing.MaxAbsScaler"

    class Inputs:
        df = Input('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's MaxAbsScaler class",
                    ["<class 'sklearn.preprocessing.data.MaxAbsScaler'>"])

    class Outputs:
        df = Output('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's MaxAbsScaler class",
                    ["<class 'sklearn.preprocessing.data.MaxAbsScaler'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                description="fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                                options = ('fit_transform','transform','inverse_transform',None))

    class Fparameters:
        # {'copy':True}
        pass

class RobustScaler(object):
    task = 'Prepare'
    subtask = 'scale'
    host = 'sklearn'
    function = 'RobustScaler'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.RobustScaler.html#sklearn.preprocessing.RobustScaler"

    class Inputs:
        df = Input('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's RobustScaler class",
                    ["<class 'sklearn.preprocessing.data.RobustScaler'>"])

    class Outputs:
        df = Output('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's RobustScaler class",
                    ["<class 'sklearn.preprocessing.data.RobustScaler'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                description="fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                                options = ('fit_transform','transform','inverse_transform',None))

    class Fparameters:
        # {'with_centering': True, 'quantile_range': (25.0, 75.0), 'copy': True, 'with_scaling': True}
        pass

class Normalizer(object):
    task = 'Prepare'
    subtask = 'scale'
    host = 'sklearn'
    function = 'Normalizer'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Normalizer.html#sklearn.preprocessing.Normalizer"

    class Inputs:
        df = Input('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's Normalizer class",
                    ["<class 'sklearn.preprocessing.data.Normalizer'>"])

    class Outputs:
        df = Output('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's Normalizer class",
                    ["<class 'sklearn.preprocessing.data.Normalizer'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                description="fit_transform: always make a new api; transform: must receive an api None: only make a new api ",
                                options = ('fit_transform','transform',None))

    class Fparameters:
        # {'copy': True, 'norm': 'l2'}
        pass

class PCA(object):
    task = 'Prepare'
    subtask = 'feature transformation'
    host = 'sklearn'
    function = 'PCA'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html#sklearn.decomposition.PCA"

    class Inputs:
        df = Input('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's PCA class",
                    ["<class 'sklearn.decomposition.pca.PCA'>"])

    class Outputs:
        df = Output('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's PCA class",
                    ["<class 'sklearn.decomposition.pca.PCA'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                description="fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                                options = ('fit_transform','transform','inverse_transform',None))


    class Fparameters:
        # {'copy': True, 'norm': 'l2'}
        pass

class KernelPCA(object):
    task = 'Prepare'
    subtask = 'feature transformation'
    host = 'sklearn'
    function = 'KernelPCA'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.KernelPCA.html#sklearn.decomposition.KernelPCA"

    class Inputs:
        df = Input('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's KernelPCA class",
                    ["<class 'sklearn.decomposition.kernel_pca.KernelPCA'>"])

    class Outputs:
        df = Output('df', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's KernelPCA class",
                    ["<class 'sklearn.decomposition.kernel_pca.KernelPCA'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                description="fit_transform: always make a new api; transform: must receive an api; inverse_transform: must receive an api; None: only make a new api ",
                                options = ('fit_transform','transform','inverse_transform',None))

    class Fparameters:
        # {'fit_inverse_transform': False, 'kernel': 'linear', 'n_jobs': 1, 'tol': 0, 'degree': 3, 'random_state': None, 'max_iter': None, 'kernel_params': None, 'remove_zero_eig': False, 'n_components': None, 'eigen_solver': 'auto', 'copy_X': True, 'alpha': 1.0, 'coef0': 1, 'gamma': None}
        pass

class train_test_split(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'train_test_split'
    module = ('sklearn','model_selection')
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])

    class Outputs:
        dfx_train = Output('dfx_train', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfx_test = Output('dfx_test', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy_train = Output('dfy_train', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy_test = Output('dfy_test', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])

    class WParameters:
        pass

    class Fparameters:
        #
        pass

class KFold(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'KFold'
    module = ('sklearn','model_selection')
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.KFold.html"

    class Inputs:
        df = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])

    class Outputs:
        fold_gen = Output('cv_gen','Generator of indices to split data into training and test set',
                        ["<type 'generator'>"])
        api = Output('api', "instance of scikit-learn's KFold class",
                    ["<class 'sklearn.model_selection._split.KFold'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('split',None))

    class Fparameters:
        #n_splits=3, shuffle=False, random_state=None
        pass

class ShuffleSplit(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'ShuffleSplit'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.ShuffleSplit.html#sklearn.model_selection.ShuffleSplit"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's ShuffleSplit class",
                    ["<class 'sklearn.model_selection._split.ShuffleSplit'>"])

    class Outputs:
        dfx_train = Output('dfx_train', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfx_test = Output('dfx_test', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy_train = Output('dfy_train', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy_test = Output('dfy_test', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's ShuffleSplit class",
                    ["<class 'sklearn.model_selection._split.ShuffleSplit'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('split',None))

    class Fparameters:
        # n_splits=10, test_size='default', train_size=None, random_state=None
        pass

class StratifiedShuffleSplit(object):
    task = 'Prepare'
    subtask = 'split'
    host = 'sklearn'
    function = 'StratifiedShuffleSplit'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.StratifiedShuffleSplit.html#sklearn.model_selection.StratifiedShuffleSplit"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's StratifiedShuffleSplit class",
                    ["<class 'sklearn.model_selection._split.StratifiedShuffleSplit'>"])

    class Outputs:
        dfx_train = Output('dfx_train', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfx_test = Output('dfx_test', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy_train = Output('dfy_train', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy_test = Output('dfy_test', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's StratifiedShuffleSplit class",
                    ["<class 'sklearn.model_selection._split.StratifiedShuffleSplit'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('split',None))

    class Fparameters:
        # n_splits=10, test_size='default', train_size=None, random_state=None
        pass

class LinearRegression(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'LinearRegression'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's LinearRegression class",
                    ["<class 'sklearn.linear_model.base.LinearRegression'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's LinearRegression class",
                    ["<class 'sklearn.linear_model.base.LinearRegression'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        # {'fit_intercept': True, 'normalize': False, 'n_jobs': 1, 'copy_X': True}
        pass

class Ridge(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'Ridge'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's Ridge class",
                    ["<class 'sklearn.linear_model.ridge.Ridge'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's Ridge class",
                    ["<class 'sklearn.linear_model.ridge.Ridge'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        # {'max_iter': None, 'normalize': False, 'tol': 0.001, 'alpha': 1.0, 'fit_intercept': True, 'copy_X': True, 'solver': 'auto', 'random_state': None}
        pass

class KernelRidge(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'KernelRidge'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.kernel_ridge.KernelRidge.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's KernelRidge class",
                    ["<class 'sklearn.kernel_ridge.KernelRidge'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's KernelRidge class",
                    ["<class 'sklearn.kernel_ridge.KernelRidge'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'gamma': None, 'coef0': 1, 'kernel_params': None, 'alpha': 1, 'degree': 3, 'kernel': 'linear'}
        pass

class Lasso(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'Lasso'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lasso.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's Lasso class",
                    ["<class 'sklearn.linear_model.coordinate_descent.Lasso'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's Lasso class",
                    ["<class 'sklearn.linear_model.coordinate_descent.Lasso'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'positive': False, 'max_iter': 1000, 'normalize': False, 'tol': 0.0001, 'selection': 'cyclic', 'alpha': 1.0,
        # 'fit_intercept': True, 'copy_X': True, 'precompute': False, 'warm_start': False, 'random_state': None}
        pass

class MultiTaskLasso(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'MultiTaskLasso'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.MultiTaskLasso.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's MultiTaskLasso class",
                    ["<class 'sklearn.linear_model.coordinate_descent.MultiTaskLasso'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's MultiTaskLasso class",
                    ["<class 'sklearn.linear_model.coordinate_descent.MultiTaskLasso'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'max_iter': 1000, 'normalize': False, 'tol': 0.0001, 'selection': 'cyclic', 'alpha': 1.0, 'fit_intercept': True,
        #  'copy_X': True, 'warm_start': False, 'random_state': None}
        pass

class ElasticNet(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'ElasticNet'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ElasticNet.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's ElasticNet class",
                    ["<class 'sklearn.linear_model.coordinate_descent.ElasticNet'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's ElasticNet class",
                    ["<class 'sklearn.linear_model.coordinate_descent.ElasticNet'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'copy_X': True, 'tol': 0.0001, 'selection': 'cyclic', 'alpha': 1.0, 'fit_intercept': True, 'normalize': False,
        #  'random_state': None, 'max_iter': 1000, 'l1_ratio': 0.5, 'positive': False, 'precompute': False, 'warm_start': False}
        pass

class MultiTaskElasticNet(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'MultiTaskElasticNet'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.MultiTaskElasticNet.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's MultiTaskElasticNet class",
                    ["<class 'sklearn.linear_model.coordinate_descent.MultiTaskElasticNet'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's MultiTaskElasticNet class",
                    ["<class 'sklearn.linear_model.coordinate_descent.MultiTaskElasticNet'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'max_iter': 1000, 'copy_X': True, 'selection': 'cyclic', 'l1_ratio': 0.5, 'alpha': 1.0, 'fit_intercept': True, 'normalize': False, 'warm_start': False, 'random_state': None, 'tol': 0.0001}
        pass

class Lars(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'Lars'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lars.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's Lars class",
                    ["<class 'sklearn.linear_model.least_angle.Lars'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's Lars class",
                    ["<class 'sklearn.linear_model.least_angle.Lars'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'eps': 2.2204460492503131e-16, 'verbose': False, 'copy_X': True, 'positive': False, 'fit_intercept': True, 'normalize': True, 'precompute': 'auto', 'n_nonzero_coefs': 500, 'fit_path': True}
        pass

class LassoLars(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'LassoLars'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoLars.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's LassoLars class",
                    ["<class 'sklearn.linear_model.least_angle.LassoLars'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's LassoLars class",
                    ["<class 'sklearn.linear_model.least_angle.LassoLars'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'eps': 2.2204460492503131e-16, 'max_iter': 500, 'copy_X': True, 'alpha': 1.0, 'fit_intercept': True,
        # 'normalize': True, 'precompute': 'auto', 'verbose': False, 'fit_path': True, 'positive': False}
        pass

class BayesianRidge(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'BayesianRidge'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.BayesianRidge.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's BayesianRidge class",
                    ["<class 'sklearn.linear_model.bayes.BayesianRidge'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's BayesianRidge class",
                    ["<class 'sklearn.linear_model.bayes.BayesianRidge'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'alpha_1': 1e-06, 'n_iter': 300, 'lambda_1': 1e-06, 'copy_X': True, 'tol': 0.001, 'normalize': False,
        # 'compute_score': False, 'fit_intercept': True, 'alpha_2': 1e-06, 'verbose': False, 'lambda_2': 1e-06}
        pass

class ARDRegression(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'ARDRegression'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ARDRegression.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's ARDRegression class",
                    ["<class 'sklearn.linear_model.bayes.ARDRegression'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's ARDRegression class",
                    ["<class 'sklearn.linear_model.bayes.ARDRegression'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'alpha_2': 1e-06, 'copy_X': True, 'tol': 0.001, 'fit_intercept': True, 'verbose': False, 'alpha_1': 1e-06,
        # 'lambda_1': 1e-06, 'threshold_lambda': 10000.0, 'compute_score': False, 'n_iter': 300, 'normalize': False, 'lambda_2': 1e-06}
        pass

class LogisticRegression(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'LogisticRegression'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's LogisticRegression class",
                    ["<class 'sklearn.linear_model.logistic.LogisticRegression'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's LogisticRegression class",
                    ["<class 'sklearn.linear_model.logistic.LogisticRegression'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'class_weight': None, 'tol': 0.0001, 'dual': False, 'fit_intercept': True, 'solver': 'liblinear', 'random_state': None, 'multi_class': 'ovr', 'C': 1.0, 'max_iter': 100, 'verbose': 0, 'penalty': 'l2', 'n_jobs': 1, 'warm_start': False, 'intercept_scaling': 1}
        pass

class SGDRegressor(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'SGDRegressor'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.SGDRegressor.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's SGDRegressor class",
                    ["<class 'sklearn.linear_model.stochastic_gradient.SGDRegressor'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's SGDRegressor class",
                    ["<class 'sklearn.linear_model.stochastic_gradient.SGDRegressor'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'shuffle': True, 'loss': 'squared_loss', 'learning_rate': 'invscaling', 'alpha': 0.0001, 'fit_intercept': True,
        #  'epsilon': 0.1, 'random_state': None, 'penalty': 'l2', 'eta0': 0.01, 'l1_ratio': 0.15, 'verbose': 0, 'n_iter': 5, 'average': False, 'warm_start': False, 'power_t': 0.25}
        pass

class SVR(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'SVR'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.svm.SVR.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's SVR class",
                    ["<class 'sklearn.svm.classes.SVR'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's SVR class",
                    ["<class 'sklearn.svm.classes.SVR'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'gamma': 'auto', 'coef0': 0.0, 'max_iter': -1, 'tol': 0.001, 'cache_size': 200, 'degree': 3, 'verbose': False,
        #  'C': 1.0, 'epsilon': 0.1, 'shrinking': True, 'kernel': 'rbf'}
        pass

class NuSVR(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'NuSVR'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.svm.NuSVR.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's NuSVR class",
                    ["<class 'sklearn.svm.classes.NuSVR'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's NuSVR class",
                    ["<class 'sklearn.svm.classes.NuSVR'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'gamma': 'auto', 'coef0': 0.0, 'max_iter': -1, 'tol': 0.001, 'cache_size': 200, 'degree': 3, 'verbose': False,
        #  'C': 1.0, 'epsilon': 0.1, 'shrinking': True, 'kernel': 'rbf'}
        pass

class LinearSVR(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'LinearSVR'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.svm.LinearSVR.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's LinearSVR class",
                    ["<class 'sklearn.svm.classes.LinearSVR'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's LinearSVR class",
                    ["<class 'sklearn.svm.classes.LinearSVR'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'fit_intercept': True, 'max_iter': 1000, 'loss': 'epsilon_insensitive', 'tol': 0.0001, 'dual': True, 'C': 1.0,
        #  'verbose': 0, 'epsilon': 0.0, 'intercept_scaling': 1.0, 'random_state': None}
        pass

class MLPRegressor(object):
    task = 'Model'
    subtask = 'regression'
    host = 'sklearn'
    function = 'MLPRegressor'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPRegressor.html#sklearn.neural_network.MLPRegressor"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Input('api', "instance of scikit-learn's MLPRegressor class",
                    ["<class 'sklearn.neural_network.multilayer_perceptron.MLPRegressor'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's MLPRegressor class",
                    ["<class 'sklearn.neural_network.multilayer_perceptron.MLPRegressor'>"])

    class WParameters:
        func_method = Parameter('func_method', None, 'string',
                                options = ('fit','predict',None))

    class Fparameters:
        #{'beta_1': 0.9, 'warm_start': False, 'beta_2': 0.999, 'shuffle': True, 'verbose': False, 'nesterovs_momentum': True,
        #  'hidden_layer_sizes': (100,), 'epsilon': 1e-08, 'activation': 'relu', 'max_iter': 200, 'batch_size': 'auto',
        # 'power_t': 0.5, 'random_state': None, 'learning_rate_init': 0.001, 'tol': 0.0001, 'validation_fraction': 0.1,
        #  'alpha': 0.0001, 'solver': 'adam', 'momentum': 0.9, 'learning_rate': 'constant', 'early_stopping': False}
        pass

class GridSearchCV(object):
    task = 'Search'
    subtask = 'grid'
    host = 'sklearn'
    function = 'GridSearchCV'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        estimator = Input('estimator', "instance of a machine learning class", regression_types())
        scorer = Input('scorer', "instance of scikit-learn's make_scorer class",
                    ["<class 'sklearn.metrics.scorer._PredictScorer'>"])

    class Outputs:
        cv_results_ = Output('cv_results_', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        api = Output('api', "instance of scikit-learn's GridSearchCV class",
                    ["<class 'sklearn.grid_search.GridSearchCV'>"])
        best_estimator_ = Output('best_estimator_', "instance of a machine learning class",regression_types())

    class WParameters:
        pass

    class Fparameters:
        #estimator, param_grid, scoring=None, fit_params=None, n_jobs=1, iid=True, refit=True, cv=None, verbose=0, pre_dispatch='2*n_jobs', error_score='raise', return_train_score=True
        pass

class cross_val_score(object):
    task = 'Search'
    subtask = 'validate'
    host = 'sklearn'
    function = 'cross_val_score'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.cross_val_score.html#sklearn.model_selection.cross_val_score"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        estimator = Input('estimator', "instance of a machine learning class", regression_types())
        scorer = Input('scorer', "instance of scikit-learn's make_scorer class",
                    ["<class 'sklearn.metrics.scorer._PredictScorer'>"])

    class Outputs:
        scores = Output('scores', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])

    class WParameters:
        pass

    class Fparameters:
        # estimator, X, y=None, groups=None, scoring=None, cv=None, n_jobs=1, verbose=0, fit_params=None, pre_dispatch='2*n_jobs'
        pass

class cross_val_predict(object):
    task = 'Search'
    subtask = 'validate'
    host = 'sklearn'
    function = 'cross_val_predict'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.cross_val_predict.html#sklearn.model_selection.cross_val_predict"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        estimator = Input('estimator', "instance of a machine learning class", regression_types())
        scorer = Input('scorer', "instance of scikit-learn's make_scorer class",
                    ["<class 'sklearn.metrics.scorer._PredictScorer'>"])

    class Outputs:
        dfy_predict = Output('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])

    class WParameters:
        pass

    class Fparameters:
        # estimator, X, y=None, groups=None, cv=None, n_jobs=1, verbose=0, fit_params=None, pre_dispatch='2*n_jobs', method='predict'
        pass

class learning_curve(object):
    task = 'Search'
    subtask = 'grid'
    host = 'sklearn'
    function = 'learning_curve'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.learning_curve.html#sklearn.model_selection.learning_curve"

    class Inputs:
        dfx = Input('dfx', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        estimator = Input('estimator', "instance of a machine learning class", regression_types())
        scorer = Input('scorer', "instance of scikit-learn's make_scorer class",
                    ["<class 'sklearn.metrics.scorer._PredictScorer'>"])
        cv = Input('cv', "instance of scikit-learn's cross validation generator", cv_types())

    class Outputs:
        train_sizes_abs = Output('train_sizes_abs', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        train_scores = Output('train_scores', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        test_scores = Output('test_scores', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        extended_result_ = Output('extended_result_', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])

    class WParameters:
        pass

    class Fparameters:
        # estimator, X, y, groups=None, train_sizes=array([ 0.1, 0.33, 0.55, 0.78, 1. ]), cv=None, scoring=None, exploit_incremental_learning=False, n_jobs=1, pre_dispatch='all', verbose=0, shuffle=False, random_state=None
        pass

class scorer_regression(object):
    task = 'Search'
    subtask = 'evaluate'
    host = 'sklearn'
    function = 'scorer_regression'
    requirements = (req(0))
    documentation = "http://scikit-learn.org/0.15/modules/generated/sklearn.metrics.make_scorer.html#sklearn.metrics.make_scorer"

    class Inputs:
        pass

    class Outputs:
        scorer = Output('scorer', 'Callable object that returns a scalar score',
                        ["<class 'sklearn.metrics.scorer._PredictScorer'>"])

    class WParameters:
        metric = Parameter('metric', 'mae', "http://scikit-learn.org/dev/modules/model_evaluation.html#regression-metrics",
                                options = ('mae','mse','rmse','r2'))

    class Fparameters:
        #score_func, greater_is_better=True, needs_proba=False, needs_threshold=False, **kwargs
        pass

class evaluate_regression(object):
    task = 'Search'
    subtask = 'evaluate'
    host = 'sklearn'
    function = 'evaluate_regression'
    requirements = (req(0), req(2))
    documentation = "http://scikit-learn.org/dev/modules/model_evaluation.html#regression-metrics"

    class Inputs:
        dfy_predict = Input('dfy_predict', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        dfy = Input('dfy', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])

    class Outputs:
        evaluation_results_ = Output('evaluation_results_', 'pandas dataframe', ["<class 'pandas.core.frame.DataFrame'>"])
        evaluator = Output('evaluator', "dictionary of metrics and their score function",["<type 'dict'>"])

    class WParameters:
        r2_score = Parameter('r2_score', False,
                           "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html#sklearn.metrics.r2_score",
                           options=(True, False))
        r2_sample_weight = Parameter('r2_sample_weight', None,
                         "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html#sklearn.metrics.r2_score")
        multioutput = Parameter('multioutput', 'uniform_average',
                         "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html#sklearn.metrics.r2_score",
                         options=('raw_values', 'uniform_average', 'variance_weighted'))

        mean_absolute_error = Parameter('mean_absolute_error', False,
                             "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html#sklearn.metrics.mean_absolute_error",
                             options=(True, False))
        mae_sample_weight = Parameter('mae_sample_weight', None,
                                     "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html#sklearn.metrics.mean_absolute_error")
        mae_multioutput = Parameter('mae_multioutput', 'uniform_average',
                                "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html#sklearn.metrics.mean_absolute_error",
                                options=('raw_values', 'uniform_average'))

        median_absolute_error = Parameter('median_absolute_error', False,
                                    "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.median_absolute_error.html#sklearn.metrics.median_absolute_error",
                                    options=(True, False))

        mean_squared_error = Parameter('mean_squared_error', False,
                                      "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error",
                                      options=(True, False))
        mse_sample_weight = Parameter('mse_sample_weight', None,
                                          "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error")
        mse_multioutput = Parameter('mse_multioutput', 'uniform_average',
                                          "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error",
                                        options=('raw_values', 'uniform_average'))

        root_mean_squared_error = Parameter('root_mean_squared_error', False,
                                   "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error",
                                   options=(True, False))
        rmse_sample_weight = Parameter('rmse_sample_weight', None,
                                  "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error")
        rmse_multioutput = Parameter('rmse_multioutput', 'uniform_average',
                                "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html#sklearn.metrics.mean_squared_error",
                                options=('raw_values', 'uniform_average'))

        explained_variance_score = Parameter('explained_variance_score', False,
                                        "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.explained_variance_score.html#sklearn.metrics.explained_variance_score",
                                        options=(True, False))
        ev_sample_weight = Parameter('ev_sample_weight', None,
                                   "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.explained_variance_score.html#sklearn.metrics.explained_variance_score")
        ev_multioutput = Parameter('ev_multioutput', 'uniform_average',
                                 "http://scikit-learn.org/stable/modules/generated/sklearn.metrics.explained_variance_score.html#sklearn.metrics.explained_variance_score",
                                 options=('raw_values', 'uniform_average', 'variance_weighted'))

    class Fparameters:
        pass




# [klass[0] for klass in inspect.getmembers(sklearn_db)] =
#['ARDRegression', 'BayesianRidge', 'Binarizer', 'ElasticNet', 'GridSearchCV', 'Imputer', 'Input', 'KFold', 'KernelPCA', 'KernelRidge', 'Lars', 'Lasso', 'LassoLars', 'LinearRegression', 'LinearSVR', 'LogisticRegression', 'MLPRegressor', 'MaxAbsScaler', 'MinMaxScaler', 'MultiTaskElasticNet', 'MultiTaskLasso', 'Normalizer', 'NuSVR', 'OneHotEncoder', 'Output', 'PCA', 'Parameter', 'PolynomialFeatures', 'Ridge', 'RobustScaler', 'SGDRegressor', 'SVR', 'ShuffleSplit', 'StandardScaler', 'StratifiedShuffleSplit', '__builtins__', '__doc__', '__file__', '__name__', '__package__', 'cross_val_predict', 'cross_val_score', 'cv_types', 'evaluate_regression', 'learning_curve', 'mask', 'np', 'regression_types', 'req', 'scorer_regression', 'train_test_split']
# len = all + 5 '__' + 6 imported + 1 np + 1 mask


import unittest
from cheml.initialization import XYZreader
from cheml.initialization import ConvertFile


class TestConvertFile(unittest.TestCase):
    def test_xyz_cml(self):
        reader= XYZreader(path_pattern='[2-3]/*.opt.xyz',
                                           path_root = 'data/',
                                           reader='manual',
                                           skip_lines=[0, 0])
        molecules = reader.read()
        model = ConvertFile(molecules, 'xyz', 'cml')
        # commented out since this command generates some files
        # s = model.convert()
        # self.assertEqual(len(s) , 2)
        # self.assertEqual(s[1]['file'][-3:], 'cml')

if __name__== '__main__':
    unittest.main()






# coding=utf-8
import math
import numpy as np
from ....data.materials.util.LookUpData import LookUpData
from ....vassal.analysis.PairDistanceAnalysis import PairDistanceAnalysis

class PRDFRegression:
    """Class that uses the partial radial distribution function to perform
    regression.

    Each material is represented by a matrix containing the PRDF
    between each element type. Distance between structures are computed as
    Frobenius norm of the difference between their PRDF matrices. If you use
    this method, please cite Schutt, Glawe, et al. [1].

    Attributes
    ----------
    sigma : float
        Normalization term in kernel function.
    n_bins : int
        Number of distance points to evaluate.
    cut_off_distance : float
        Cutoff distance for PRDF.

    References
    ----------
    .. [1] K. T. Schütt, H. Glawe, F. Brockherde, A. Sanna, K. R. Müller,
    and E. K. U. Gross, "How to represent crystal structures for machine
    learning: Towards fast prediction of electronic properties," Physical
    Review B, vol. 89, no. 20, May 2014.

    """

    def __init__(self):
        """Function to create instance and initialize fields.
        """

        # Normalization term in kernel function.
        self.sigma = 1

        # Number of bins in PRDF.
        self.n_bins = 25

        # Cutoff distance of PRDF.
        self.cut_off = 7.0

    def set_sigma(self, s):
        """Function to set the normalization parameter in the kernel function.

        Parameters
        ----------
        s : float
            Desired normalization parameter.

        """
        self.sigma = s

    def set_n_bins(self, n_b):
        """Function to set the number of bins used when computing the PRDF.

        Parameters
        ----------
        n_b : int
            Number of bins (>1)

        Raises
        ------
        ValueError
            If n_bins is less than 1.

        """
        if n_b < 1:
            raise ValueError("# bins must be greater than 1.")
        self.n_bins = n_b

    def set_cut_off(self, d):
        """Function to set the cutoff distance used when computing the PRDF.

        Parameters
        ----------
        d : float
            Cutoff distance (>0 Angstrom)

        Raises
        ------
        ValueError
            If d is less than or equal to 0.
        """
        if d <= 0:
            raise ValueError("Distance must be positive.")
        self.cut_off = d

    def compute_similarity(self, structure1, structure2):
        """Function to compute similarity between two crystal structures.

        Parameters
        ----------
        structure1 : dict
            Representation of structure #1. Dictionary containing a tuple of
            integers as the key and a list of floats as the values.
        structure2 : dict
            Representation of structure #2. Dictionary containing a tuple of
            integers as the key and a list of floats as the values.

        Returns
        -------
        output : float
            Similarity between the two structures.

        """
        pairs = set()

        # Compile a complete list of the element pairs for both structures.
        for k in structure1.keys():
            pairs.add(k)
        for k in structure2.keys():
            pairs.add(k)

        # For each pair, compute the squared differences between the two PRDFs.
        # This is equivalent to the Froebius norm.
        diff = 0.0
        for pair in pairs:
            if pair not in structure1:
                # Assume prdf1 == 0.
                diff += sum(structure2[pair] ** 2)
            elif pair not in structure2:
                # Assume prdf2 == 0.
                diff += sum(structure1[pair] ** 2)
            else:
                diff += sum((structure1[pair] - structure2[pair]) ** 2)

        # Compute kernel function to get similarity.
        return math.exp(-1 * diff / self.sigma)

    def compute_representation(self, structure):
        """Function to compute the representation of the crystal structure.

        Parameters
        ----------
        structure : CrystalStructureEntry
            Crystal structure.

        Returns
        -------
        output : dict
            Representation of the structure. Dictionary containing a tuple of
            integers as the key and a list of floats as the values.

        Raises
        ------
        Exception
            If element name is invalid.
        RuntimeError
            If something goes wrong while analyzing structure.
        """

        # Get the atomic number of each type.
        n_types = structure.n_types()
        type_z = np.zeros(n_types, dtype=int)
        for i in range(n_types):
            if structure.get_type_name(i) in LookUpData.element_names:
                type_z[i] = LookUpData.element_names.index(
                    structure.get_type_name(i))
            else:
                raise Exception(
                    "No such element: " + structure.get_type_name(i))

        # Compute the PRDF of this structure.
        pda = PairDistanceAnalysis()
        pda.set_cutoff_distance(self.cut_off)

        try:
            pda.analyze_structure(structure)
        except Exception:
            raise RuntimeError("Oops, something went wrong in analyzing "
                               "structure.")

        prdf = pda.compute_PRDF(self.n_bins)
        rep = {}
        for i in range(len(prdf)):
            for j in range(len(prdf[i])):
                rep[(type_z[i], type_z[j])] = prdf[i][j]

        return rep


import unittest
import numpy as np
import math
from ....vassal.analysis.VoronoiCellBasedAnalysis import \
    VoronoiCellBasedAnalysis
from ....vassal.data.Atom import Atom
from ....vassal.data.Cell import Cell
import numpy.testing as np_tst

class testVoronoiCellBasedAnalysis(unittest.TestCase):
    def test_BCC(self):
        # Structure of bcc.
        structure = Cell()
        atom = Atom([0, 0, 0], 0)
        structure.add_atom(atom)
        atom = Atom([0.5, 0.5, 0.5], 0)
        structure.add_atom(atom)

        # Prepare.
        tool = VoronoiCellBasedAnalysis(radical=True)
        tool.analyze_structure(structure)

        # Check results.
        n_eff = 11.95692194
        np_tst.assert_array_almost_equal([n_eff, n_eff],
                            tool.get_effective_coordination_numbers())
        self.assertAlmostEquals(14.0, tool.face_count_average(), delta=1e-2)
        self.assertAlmostEquals(0.0, tool.face_count_variance(), delta=1e-2)
        self.assertAlmostEquals(14.0, tool.face_count_minimum(), delta=1e-2)
        self.assertAlmostEquals(14.0, tool.face_count_maximum(), delta=1e-2)
        self.assertAlmostEquals(1, len(tool.get_unique_polyhedron_shapes()),
                                delta=1e-2)
        self.assertAlmostEquals(0.0, tool.volume_variance(), delta=1e-2)
        self.assertAlmostEquals(0.5, tool.volume_fraction_minimum(),
                                delta=1e-2)
        self.assertAlmostEquals(0.5, tool.volume_fraction_maximum(),
                                delta=1e-2)
        self.assertAlmostEquals(0.68, tool.max_packing_efficiency(),
                                delta=1e-2)
        self.assertAlmostEquals(0, tool.mean_bcc_dissimilarity(),
                                delta=1e-2)
        self.assertAlmostEquals(14.0 / 12.0, tool.mean_fcc_dissimilarity(),
                                delta=1e-2)
        self.assertAlmostEquals(8.0 / 6.0, tool.mean_sc_dissimilarity(),
                                delta=1e-2)
        bond_lengths = tool.bond_lengths()
        self.assertEquals(2, len(bond_lengths))
        self.assertEquals(14, len(bond_lengths[0]))
        self.assertAlmostEquals(math.sqrt(3) / 2 , bond_lengths[0][0],
                                delta=1e-6)
        self.assertAlmostEquals(1.0, bond_lengths[0][12], delta=1e-6)
        mean_bond_lengths = tool.mean_bond_lengths()
        var_bond_lengths = tool.bond_length_variance(mean_bond_lengths)
        self.assertTrue(var_bond_lengths[0] > 0)

    def test_B2(self):
        # Structure of bcc.
        structure = Cell()
        atom = Atom([0, 0, 0], 0)
        structure.add_atom(atom)
        atom = Atom([0.5, 0.5, 0.5], 1)
        structure.add_atom(atom)

        # Prepare.
        tool = VoronoiCellBasedAnalysis(radical=True)
        tool.analyze_structure(structure)

        # Check results.
        n_eff = 11.95692194
        np_tst.assert_array_almost_equal([n_eff, n_eff],
                            tool.get_effective_coordination_numbers())
        self.assertAlmostEquals(14.0, tool.face_count_average(), delta=1e-2)
        self.assertAlmostEquals(0.0, tool.face_count_variance(), delta=1e-2)
        self.assertAlmostEquals(14.0, tool.face_count_minimum(), delta=1e-2)
        self.assertAlmostEquals(14.0, tool.face_count_maximum(), delta=1e-2)
        self.assertAlmostEquals(1, len(tool.get_unique_polyhedron_shapes()),
                                delta=1e-2)
        self.assertAlmostEquals(0.0, tool.volume_variance(), delta=1e-2)
        self.assertAlmostEquals(0.5, tool.volume_fraction_minimum(),
                                delta=1e-2)
        self.assertAlmostEquals(0.5, tool.volume_fraction_maximum(),
                                delta=1e-2)
        self.assertAlmostEquals(0.68, tool.max_packing_efficiency(),
                                delta=1e-2)
        self.assertAlmostEquals(0, tool.mean_bcc_dissimilarity(),
                                delta=1e-2)
        self.assertAlmostEquals(14.0 / 12.0, tool.mean_fcc_dissimilarity(),
                                delta=1e-2)
        self.assertAlmostEquals(8.0 / 6.0, tool.mean_sc_dissimilarity(),
                                delta=1e-2)
        bond_lengths = tool.bond_lengths()
        self.assertEquals(2, len(bond_lengths))
        self.assertEquals(14, len(bond_lengths[0]))
        self.assertAlmostEquals(math.sqrt(3) / 2 , bond_lengths[0][0],
                                delta=1e-6)
        self.assertAlmostEquals(1.0, bond_lengths[0][12], delta=1e-6)
        mean_bond_lengths = tool.mean_bond_lengths()
        var_bond_lengths = tool.bond_length_variance(mean_bond_lengths)
        self.assertTrue(var_bond_lengths[0] > 0)

        # Check ordering parameters (against values computed by hand).
        np_tst.assert_array_almost_equal([0.142857, -0.142857],
                                         tool.get_neighbor_ordering_parameters(
                                             1, False)[0], decimal=2)
        np_tst.assert_array_almost_equal([-0.04, 0.04],
                                         tool.get_neighbor_ordering_parameters(
                                             2, False)[0], decimal=2)
        np_tst.assert_array_almost_equal([0.551982, -0.551982],
                                         tool.get_neighbor_ordering_parameters(
                                             1, True)[0], decimal=2)
        np_tst.assert_array_almost_equal([-0.253856, 0.253856],
                                         tool.get_neighbor_ordering_parameters(
                                             2, True)[0], decimal=2)
        self.assertAlmostEquals(0.142857,
                                tool.warren_cowley_ordering_magnitude(1,
                                                                      False),
                                delta=1e-2)
        self.assertAlmostEquals(0.04,
                                tool.warren_cowley_ordering_magnitude(2,
                                                                      False),
                                delta=1e-2)
        self.assertAlmostEquals(0.551982,
                                tool.warren_cowley_ordering_magnitude(1,
                                                                      True),
                                delta=1e-2)
        self.assertAlmostEquals(0.253856,
                                tool.warren_cowley_ordering_magnitude(2,
                                                                      True),
                                delta=1e-2)

    def test_B1(self):
        # Structure of rocksalt.
        structure = Cell()
        basis = np.zeros((3, 3))
        basis[0] = np.array([0, 0.5, 0.5])
        basis[1] = np.array([0.5, 0, 0.5])
        basis[2] = np.array([0.5, 0.5, 0])
        structure.set_basis(basis=basis)
        atom = Atom([0, 0, 0], 0)
        structure.add_atom(atom)
        atom = Atom([0.5, 0.5, 0.5], 1)
        structure.add_atom(atom)

        # Prepare.
        tool = VoronoiCellBasedAnalysis(radical=True)
        tool.analyze_structure(structure)

        # Check results.
        n_eff = 6
        np_tst.assert_array_almost_equal([n_eff, n_eff],
                            tool.get_effective_coordination_numbers())
        self.assertAlmostEquals(6.0, tool.face_count_average(), delta=1e-2)
        self.assertAlmostEquals(0.0, tool.face_count_variance(), delta=1e-2)
        self.assertAlmostEquals(6.0, tool.face_count_minimum(), delta=1e-2)
        self.assertAlmostEquals(6.0, tool.face_count_maximum(), delta=1e-2)
        self.assertAlmostEquals(1, len(tool.get_unique_polyhedron_shapes()),
                                delta=1e-2)
        self.assertAlmostEquals(0.0, tool.volume_variance(), delta=1e-2)
        self.assertAlmostEquals(0.5, tool.volume_fraction_minimum(),
                                delta=1e-2)
        self.assertAlmostEquals(0.5, tool.volume_fraction_maximum(),
                                delta=1e-2)
        np_tst.assert_array_almost_equal([1, -1],
                                         tool.get_neighbor_ordering_parameters(
                                             1, False)[0], decimal=2)
        np_tst.assert_array_almost_equal([-1, 1],
                                         tool.get_neighbor_ordering_parameters(
                                             2, False)[0], decimal=2)
        np_tst.assert_array_almost_equal([1, -1],
                                         tool.get_neighbor_ordering_parameters(
                                             3, False)[0], decimal=2)
        np_tst.assert_array_almost_equal([-1, 1],
                                         tool.get_neighbor_ordering_parameters(
                                             2, True)[0], decimal=2)
        np_tst.assert_array_almost_equal([1, -1],
                                         tool.get_neighbor_ordering_parameters(
                                             3, True)[0], decimal=2)
        np_tst.assert_array_almost_equal([1, -1],
                                         tool.get_neighbor_ordering_parameters(
                                             3, True)[0], decimal=2)
        self.assertAlmostEquals(1,
                                tool.warren_cowley_ordering_magnitude(1,
                                                                      False),
                                delta=1e-2)
        self.assertAlmostEquals(1,
                                tool.warren_cowley_ordering_magnitude(2,
                                                                      False),
                                delta=1e-2)
        self.assertAlmostEquals(1,
                                tool.warren_cowley_ordering_magnitude(1,
                                                                      True),
                                delta=1e-2)
        self.assertAlmostEquals(1,
                                tool.warren_cowley_ordering_magnitude(2,
                                                                      True),
                                delta=1e-2)
        bond_lengths = tool.bond_lengths()
        self.assertEquals(2, len(bond_lengths))
        self.assertEquals(6, len(bond_lengths[0]))
        self.assertAlmostEquals(0.5 , bond_lengths[0][0], delta=1e-6)
        mean_bond_lengths = tool.mean_bond_lengths()
        self.assertEquals(2, len(mean_bond_lengths))
        self.assertAlmostEquals(0.5, mean_bond_lengths[0], delta=1e-6)
        var_bond_lengths = tool.bond_length_variance(mean_bond_lengths)
        self.assertAlmostEquals(0, var_bond_lengths[0], delta=1e-6)

        # Neighbor property attributes.
        np_tst.assert_array_almost_equal([1, 1],
                                         tool.neighbor_property_differences([
                                             0, 1], 1))
        np_tst.assert_array_almost_equal([0, 0],
                                         tool.neighbor_property_differences([
                                             0, 1], 2))
        np_tst.assert_array_almost_equal([0, 0],
                                         tool.neighbor_property_variances([
                                             0, 1], 1))
        np_tst.assert_array_almost_equal([0, 0],
                                         tool.neighbor_property_variances([
                                             0, 1], 2))

    def test_L12(self):
        # Structure of L12.
        structure = Cell()
        atom = Atom([0, 0, 0], 1)
        structure.add_atom(atom)
        atom = Atom([0.5, 0.5, 0], 0)
        structure.add_atom(atom)
        atom = Atom([0.5, 0, 0.5], 0)
        structure.add_atom(atom)
        atom = Atom([0, 0.5, 0.5], 0)
        structure.add_atom(atom)

        # Prepare.
        tool = VoronoiCellBasedAnalysis(radical=True)
        tool.analyze_structure(structure)

        # Check results.
        np_tst.assert_array_almost_equal([12, 12, 12, 12],
                            tool.get_effective_coordination_numbers())
        self.assertAlmostEquals(12.0, tool.face_count_average(), delta=1e-2)
        self.assertAlmostEquals(0.0, tool.face_count_variance(), delta=1e-2)
        self.assertAlmostEquals(12.0, tool.face_count_minimum(), delta=1e-2)
        self.assertAlmostEquals(12.0, tool.face_count_maximum(), delta=1e-2)
        self.assertAlmostEquals(1, len(tool.get_unique_polyhedron_shapes()),
                                delta=1e-2)
        self.assertAlmostEquals(0.0, tool.volume_variance(), delta=1e-2)
        self.assertAlmostEquals(0.25, tool.volume_fraction_minimum(),
                                delta=1e-2)
        self.assertAlmostEquals(0.25, tool.volume_fraction_maximum(),
                                delta=1e-2)
        self.assertAlmostEquals(0.74, tool.max_packing_efficiency(),
                                delta=1e-2)
        self.assertAlmostEquals(1, tool.mean_bcc_dissimilarity(),
                                delta=1e-2)
        self.assertAlmostEquals(0, tool.mean_fcc_dissimilarity(),
                                delta=1e-2)
        self.assertAlmostEquals(1, tool.mean_sc_dissimilarity(),
                                delta=1e-2)
        bond_lengths = tool.bond_lengths()
        self.assertEquals(4, len(bond_lengths))
        self.assertEquals(12, len(bond_lengths[0]))
        self.assertAlmostEquals(math.sqrt(2) / 2, bond_lengths[0][0],
                                delta=1e-6)
        mean_bond_lengths = tool.mean_bond_lengths()
        var_bond_lengths = tool.bond_length_variance(mean_bond_lengths)
        self.assertAlmostEquals(0, var_bond_lengths[0], delta=1e-6)

        # Neighbor analysis results.
        neigh_diff = tool.neighbor_property_differences([-1, 1], 1)
        np_tst.assert_array_almost_equal([2, 8.0 / 12.0, 2.0 / 3, 2.0 / 3],
                                         neigh_diff, decimal=5)
        neigh_diff = tool.neighbor_property_differences([-1, 1], 2)
        np_tst.assert_array_almost_equal([192.0 / 132, 64.0 / 132.0, 64.0 /
                                          132, 64.0/ 132], neigh_diff,
                                         decimal=5)

        neigh_var = tool.neighbor_property_variances([-1, 1], 1)
        # Type 0 has 8 type 0 NNs and 4 type 1, mean = -1/3.
        # Type 1 has 12 type 0 NNs, therefore variance = 0.
        # variance = 8/12 * (-1 + 1/3)^2 + 4/12 + (1 + 1/3)^2.
        a_var = 2.0 / 3.0 * (-1 + 1.0 / 3.0) ** 2 + 1.0 / 3.0 * (1 + 1.0 /
                                                                 3.0) ** 2
        np_tst.assert_array_almost_equal([0, a_var, a_var, a_var], neigh_var,
                                         decimal=5)
        neigh_var = tool.neighbor_property_variances([-1, 1], 2)
        # Type 0 has 68 type 0 2nd NN and 16 type 1.
        # Type 1 has 48 type 0 2nd NN and 36 type 1
        type0_var = 0.734618916
        type1_var = 0.79338843
        np_tst.assert_array_almost_equal([type1_var, type0_var, type0_var,
                                          type0_var], neigh_var, decimal=5)

    def test_constructor(self):
        # Create a Voronoi tessellation of a BCC crystal.
        bcc = Cell()
        atom = Atom([0, 0, 0], 0)
        bcc.add_atom(atom)
        atom = Atom([0.5, 0.5, 0.5], 0)
        bcc.add_atom(atom)

        # Prepare.
        tool = VoronoiCellBasedAnalysis(radical=True)
        tool.analyze_structure(bcc)

        # Create a new instance based on a B2 crystal.
        b2 = bcc.__copy__()
        b2.get_atom(1).set_type(1)
        tool2 = VoronoiCellBasedAnalysis(old_tessellation=tool,
                                         new_structure=b2)

        # Make sure it balks if I change the basis.
        b2.set_basis(lengths=[2, 1, 1], angles=[90, 90, 90])
        self.assertRaises(Exception.__class__)


import unittest
from .....data.utilities.generators.PhaseDiagramCompositionEntryGenerator \
    import PhaseDiagramCompositionEntryGenerator

class testPhaseDiagramCompositionEntryGenerator(unittest.TestCase):
    def setUp(self):
        self.pg = PhaseDiagramCompositionEntryGenerator()

    def tearDown(self):
        self.pg = None

    def test_generate_alloy_compositions(self):
        self.pg.set_even_spacing(True)
        self.pg.set_size(5)
        self.pg.set_order(1, 3)

        comps = self.pg.generate_alloy_compositions()
        self.assertEquals(1, len(comps[1]))
        self.assertEquals(3, len(comps[2]))
        self.assertEquals(3, len(comps[3]))

        self.pg.set_order(3, 3)
        comps = self.pg.generate_alloy_compositions()
        self.assertEquals(1, len(comps))

    def test_generate_crystal_compositions(self):
        self.pg.set_even_spacing(False)
        self.pg.set_size(4)
        self.pg.set_order(1, 3)

        comps = self.pg.generate_crystal_compositions()
        self.assertEquals(1, len(comps[1]))
        self.assertEquals(5, len(comps[2]))
        self.assertEquals(4, len(comps[3]))

        # Try making entries.
        self.pg.set_elements_by_index([0, 1, 2])
        entries = self.pg.generate_entries()
        self.assertEquals(3 * 1 + 3 * 5 + 4, len(entries))

        self.pg.set_elements_by_index([0, 1, 2, 3])
        entries = self.pg.generate_entries()
        self.assertEquals(4 * 1 + 6 * 5 + 4 * 4, len(entries))

        self.pg.set_order(2, 3)
        comps = self.pg.generate_crystal_compositions()
        self.assertEquals(2, len(comps))


# coding=utf-8
from ..PairDistanceAnalysis import PairDistanceAnalysis
from ..voronoi.VoronoiCell import VoronoiCell

class VoronoiTessellationCalculator:
    """Class that computes the Voronoi tessellation of a cell.

    Can either use a standard Voronoi tessellation or the radical plane
    method. The radical plane method takes the radii of atoms into account
    when partitioning the cell. Citation for Radical Voronoi method [1].

    References
    ----------
    .. [1] B. J. Gellatly and J. L. Finney, "Characterisation of models of
    multicomponent amorphous metals: The radical alternative to the Voronoi
    polyhedron," Journal of Non-Crystalline Solids, vol. 50, no. 3,
    pp. 313--329, Aug. 1982.

    """
    @classmethod
    def compute(self, cell, radical):
        """Function to run a a tessellation using the Python implementation
        of the voronoi tessellation provided with Vassal.

        Citation for the computation method [1].

        Parameters
        ----------
        cell : Cell
            Structure to analyze.
        radical : bool
            Whether to perform a radical plane tessellation.

        Returns
        -------
        output : array-like
            Voronoi cell for each atom.

        References
        ----------
        .. [1] W. Brostow, J.-P. Dussault, and B. L. Fox, "Construction of
        Voronoi polyhedra," Journal of Computational Physics, vol. 29, no. 1,
        pp. 81--92, Oct. 1978.
        """

        # Initialize Voronoi cells.
        output = [VoronoiCell(atom, radical) for atom in cell.get_atoms()]

        # Create tool to find closest images.
        image_finder = PairDistanceAnalysis()
        atom_length_scale = (cell.volume()/ cell.n_atoms()) ** (1/3.0)

        image_finder.analyze_structure(cell)

        # Generate cells.
        for c in output:
            c.compute_cell(image_finder, cutoff=atom_length_scale * 6)

        # Read the.
        return output


"""
The cheml.preprocessing module includes (please click on links adjacent to function names for more information):
    - MissingValues: :func:`~cheml.preprocessing.MissingValues`
    - ConstantColumns: :func:`~cheml.preprocessing.ConstantColumns`
    - Outliers: :func:`~cheml.preprocessing.Outliers`
"""

from .handle_missing import MissingValues

from .purge import ConstantColumns
from .purge import Outliers


__all__ = [
    'MissingValues',
    'ConstantColumns',
    'Outliers'
]



import unittest
import os
from .....data.materials.CompositionEntry import CompositionEntry
from .....data.materials.util.GCLPCalculator import GCLPCalculator
from .....data.materials.util.LookUpData import LookUpData

class testGCLPCalculator(unittest.TestCase):
    this_file_path = os.path.dirname(__file__)
    abs_path = os.path.join(this_file_path, "../../../test-files/")
    def setUp(self):
        self.calc = GCLPCalculator()

    def tearDown(self):
        self.calc = None

    def test_initialization(self):
        n_elem = self.calc.get_num_phases()
        self.assertEquals(len(LookUpData.element_names), n_elem, "Initial "
                                                                 "number"
                                        " of phases should be equal to 112")
        # Add in NaCl.
        NaCl = CompositionEntry("NaCl")
        self.calc.add_phase(NaCl, -1)
        self.assertEquals(1 + n_elem, self.calc.get_num_phases())

        # Add in a duplicate.
        self.calc.add_phase(NaCl, -1)
        self.assertEquals(1 + n_elem, self.calc.get_num_phases())

        # See if energy is updated.
        self.calc.add_phase(NaCl, 0)
        self.assertAlmostEquals(-1, self.calc.phases[NaCl], delta=1e-6)
        self.calc.add_phase(NaCl, -2)
        self.assertAlmostEquals(-2, self.calc.phases[NaCl], delta=1e-6)

        # Add many phases.
        entries = CompositionEntry.import_composition_list(
            self.abs_path+"small_set_comp.txt")
        energies = CompositionEntry.import_values_list(
            self.abs_path+"small_set_delta_e.txt")
        self.calc.add_phases(entries, energies)

        self.assertEquals(725, self.calc.get_num_phases(),
                          "Total number of phases should be equal.")

    def test_GCLP(self):
        n_elem = self.calc.get_num_phases()
        self.assertEqual(len(LookUpData.element_names), n_elem, "Initial number"
                                        " of phases should be equal to 112")

        # Simple test: No phases.
        NaCl = CompositionEntry("NaCl")
        left, right = self.calc.run_GCLP(NaCl)
        self.assertAlmostEquals(0.0, left, delta=1e-6)
        self.assertEquals(2, len(right))

        # Add in Na2Cl and NaCl2 to map.
        self.calc.add_phase(CompositionEntry("Na2Cl"), -1)
        self.calc.add_phase(CompositionEntry("NaCl2"), -1)
        left, right = self.calc.run_GCLP(NaCl)
        self.assertAlmostEquals(-1, left, delta=1e-6)
        self.assertEquals(2, len(right))

        # Add NaCl to the map.
        self.calc.add_phase(NaCl, -2)
        left, right = self.calc.run_GCLP(NaCl)
        self.assertAlmostEquals(-2, left, delta=1e-6)
        self.assertEquals(1, len(right))

        # Make sure it can do systems not included in original.
        left, right = self.calc.run_GCLP(CompositionEntry(
            "AlNiFeZrTiSiBrFOSeKHHe"))
        self.assertAlmostEquals(0.0, left, delta=1e-6)
        self.assertEquals(13, len(right))